__author__ = "Guillaume Bonamis"
__license__ = "MIT"
__copyright__ = "2015, ESRF"

import os
import sys
import numpy
import matplotlib
# matplotlib.use('Agg')
import matplotlib.pyplot as plt
from freesas.model import SASModel
import itertools
from scipy.optimize import fmin
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("log_freesas")


class InputModels:
    def __init__(self):
        self.inputfiles = []
        self.sasmodels = []
        self.rfactors = []
        self.rmax = None
        self.validmodels = []

    def __repr_(self):
        return "Preparation of %s models for alignment" % len(self.inputfiles)

    def assign_models(self, molecule=None):
        """
        Create SASModels from pdb files saved in self.inputfiles and saved them in self.models.
        Center of mass, inertia tensor and canonical parameters are computed for each SASModel.

        :param molecule: optional 2d array, coordinates of the atoms for the model to create
        :return self.models: list of SASModel
        """
        if not self.inputfiles and len(molecule) == 0:
            logger.error("No input files")

        if self.inputfiles:
            for inputpdb in self.inputfiles:
                model = SASModel()
                model.read(inputpdb)
                model.centroid()
                model.inertiatensor()
                model.canonical_parameters()
                self.sasmodels.append(model)
            if len(self.inputfiles) != len(self.sasmodels):
                logger.error("Problem of assignment\n%s models for %s files" % (len(self.sasmodels), len(self.inputfiles)))

        elif len(molecule) != 0:
            model = SASModel()
            model.atoms = molecule
            model.centroid()
            model.inertiatensor()
            model.canonical_parameters()
            self.sasmodels.append(model)

        return self.sasmodels

    def rcalculation(self):
        """
        Calculation the maximal value for the R-factors, which is the mean of all the R-factors of 
        inputs plus 2 times the standard deviation.
        R-factors are saved in the attribute self.rfactors, 1d array, and in percentage.

        :return rmax: maximal value for the R-factor 
        """
        if len(self.sasmodels) == 0:
            self.assign_models()
        models = self.sasmodels

        rfactors = numpy.empty(len(models), dtype="float")
        for i in range(len(models)):
            rfactors[i] = models[i].rfactor
        self.rfactors = 100.0 * rfactors

        rmax = self.rfactors.mean() + 2 * self.rfactors.std()
        self.rmax = rmax

        return rmax

    def models_selection(self):
        """
        Check if each model respect the limit for the R-factor

        :return self.validmodels: 1d array, 0 for a non valid model, else 1
        """
        if self.rmax is None:
            self.rcalculation()
        rmax = self.rmax

        validmodels = []
        for i in range(len(self.sasmodels)):
            rfactor = self.rfactors[i]
            if rfactor <= rmax:
                validmodels.append(1.0)
            else:
                validmodels.append(0.0)

        self.validmodels = numpy.array(validmodels, dtype="float")

        return self.validmodels

    def rfactorplot(self, filename=None, save=False):
        """
        Create a png file with the table of R factor for each model.
        A threshold is computed to discarded models with Rfactor>Rmax.

        :param filename: filename for the figure, default to Rfactor.png
        :param save: save automatically the figure if True, else show it
        :return fig: the wanted figures
        """
        if filename is None:
            filename = "Rfactor.png"
        if len(self.validmodels) == 0:
            self.models_selection()

        dammif_files = len(self.inputfiles)
        R = self.rfactors
        Rmax = self.rmax

        xticks = 1 + numpy.arange(dammif_files)
        fig = plt.figure(figsize=(7.5, 10))
        labels = [os.path.splitext(os.path.basename(self.inputfiles[i]))[0] for i in range(dammif_files)]

        ax2 = fig.add_subplot(1, 1, 1)
        ax2.set_title("Selection of dammif models based on R factor")
        ax2.bar(xticks - 0.5, R)
        ax2.plot([0.5, dammif_files + 0.5], [Rmax, Rmax], "-r", label="R$_{max}$ = %.3f" % Rmax)
        ax2.set_ylabel("R factor in percent")
        ax2.set_xticks(xticks)
        ax2.set_xticklabels(labels, rotation=90)
        ax2.legend(loc=8)

        bbox_props = dict(fc="pink", ec="r", lw=1)
        for i in range(dammif_files):
            if not self.validmodels[i]:
                ax2.text(i + 0.95, Rmax / 2, "Discarded", ha="center", va="center", rotation=90, size=10, bbox=bbox_props)
                logger.info("model %s discarded, Rfactor > Rmax" % self.inputfiles[i])

        if save:
            fig.savefig(filename)
        else:
            fig.show()

        return fig


class AlignModels:
    """
    Used to align DAM from pdb files
    """

    def __init__(self, files, slow=True, enantiomorphs=True):
        """
        :param files: list of pdb files to read to create DAM
        :param slow: optimized every symmetry if True, else only optimized the best one
        :param enantiomorphs: take into account both enantiomorphs if True (i.e. inversion authorized)
        """
        self.slow = slow
        self.enantiomorphs = enantiomorphs
        self.inputfiles = files
        self.outputfiles = []
        self.models = []
        self.arrayNSD = None
        self.validmodels = []
        self.reference = None

    def __repr__(self):
        return "alignment process for %s models" % len(self.models)

    def assign_models(self):
        """
        Create SASModels from pdb files saved in self.inputfiles and saved them in self.models.
        Center of mass, inertia tensor and canonical parameters are computed for each SASModel.

        :return self.models: list of SASModel
        """
        for inputpdb in self.inputfiles:
            model = SASModel()
            model.read(inputpdb)
            model.centroid()
            model.inertiatensor()
            model.canonical_parameters()
            self.models.append(model)
        if len(self.inputfiles) != len(self.models):
            logger.error("Problem of assignment\n%s models for %s files" % (len(self.models), len(self.inputfiles)))

        return self.models

    def optimize(self, reference, molecule, symmetry):
        """
        Use scipy.optimize to optimize transformation parameters to minimize NSD

        :param reference: SASmodel
        :param molecule: SASmodel        
        :param symmetry: 3-list of +/-1
        :return p: transformation parameters optimized
        :return dist: NSD after optimization
        """
        p, dist, niter, nfuncalls, warmflag = fmin(reference.dist_after_movement, molecule.can_param, args=(molecule, symmetry), ftol=1e-4, maxiter=200, full_output=True, disp=False)
        if niter == 200:
            logger.debug("convergence not reached")
        else:
            logger.debug("convergence reach after %s iterations" % niter)
        return p, dist

    def alignment_sym(self, reference, molecule):
        """
        Apply 8 combinations to the molecule and select the one which minimize the distance between it and the reference.

        :param reference: SASModel, the one which do not move
        :param molecule: SASModel, the one wich has to be aligned
        :return combinaison: best symmetry to minimize NSD
        :return p: transformation parameters optimized if slow is true, unoptimized else
        """
        can_paramref = reference.can_param
        can_parammol = molecule.can_param

        ref_can = reference.transform(can_paramref, [1, 1, 1])
        mol_can = molecule.transform(can_parammol, [1, 1, 1])

        if self.slow:
            parameters, dist = self.optimize(reference, molecule, [1, 1, 1])
        else:
            parameters = can_parammol
            dist = reference.dist(molecule, ref_can, mol_can)
        combinaison = None

        for comb in itertools.product((-1, 1), repeat=3):
            if comb == (1, 1, 1):
                continue
            if not self.enantiomorphs and comb[0] * comb[1] * comb[2] == -1:
                continue

            sym = numpy.diag(comb + (1,))
            mol_sym = numpy.dot(sym, mol_can.T).T

            if self.slow:
                symmetry = [sym[0, 0], sym[1, 1], sym[2, 2]]
                p, d = self.optimize(reference, molecule, symmetry)
            else:
                p = can_parammol
                d = reference.dist(molecule, ref_can, mol_sym)

            if d < dist:
                dist = d
                parameters = p
                combinaison = comb
        if combinaison is not None:
            combinaison = list(combinaison)
        else:
            combinaison = [1, 1, 1]
        return combinaison, parameters

    def makeNSDarray(self):
        """
        Calculate the NSD correlation table and save it in self.arrayNSD

        :return self.arrayNSD: 2d array, NSD correlation table
        """
        models = self.models
        size = len(models)
        valid = self.validmodels
        self.arrayNSD = numpy.zeros((size, size), dtype="float")

        for i in range(size):
            if valid[i] == 1.0:
                reference = models[i]
            else:
                self.arrayNSD[i, :] = 0.00
                continue
            for j in range(size):
                if i == j:
                    self.arrayNSD[i, j] = 0.00
                elif i < j:
                    if valid[j] == 1.0:
                        molecule = models[j]
                        symmetry, p = self.alignment_sym(reference, molecule)
                        if self.slow:
                            dist = reference.dist_after_movement(p, molecule, symmetry)
                        else:
                            p, dist = self.optimize(reference, molecule, symmetry)
                    else:
                        dist = 0.00
                    self.arrayNSD[i, j] = self.arrayNSD[j, i] = dist
        return self.arrayNSD

    def plotNSDarray(self, rmax=None, filename=None, save=False):
        """
        Create a png file with the table of NSD and the average NSD for each model.
        A threshold is computed to segregate good models and the ones to exclude.

        :param rmax: threshold of R factor for the validity of a model
        :param filename: filename for the figure, default to nsd.png
        :param save: save automatically the figure if True, else show it
        :return fig: the wanted figures
        """
        if self.arrayNSD is None:
            self.makeNSDarray()
        if not self.reference:
            self.reference = self.find_reference()
        if filename is None:
            filename = "nsd.png"

        dammif_files = len(self.inputfiles)
        valid_models = self.validmodels
        labels = [os.path.splitext(os.path.basename(self.outputfiles[i]))[0] for i in range(dammif_files)]
        mask2d = (numpy.outer(valid_models, valid_models))
        tableNSD = self.arrayNSD * mask2d
        maskedNSD = numpy.ma.masked_array(tableNSD, mask=numpy.logical_not(mask2d))
        data = valid_models * (tableNSD.sum(axis=-1) / (valid_models.sum() - 1))  # mean for the valid models, excluding itself
        
        fig = plt.figure(figsize=(15, 10))
        xticks = 1 + numpy.arange(dammif_files)
        ax1 = fig.add_subplot(1, 2, 1)
        ax2 = fig.add_subplot(1, 2, 2)

        # first subplot : the NSD table
        lnsd = []
        for i in range(dammif_files):
            for j in range(dammif_files):
                nsd = maskedNSD[i, j]
                if not maskedNSD.mask[i, j]:
                    ax1.text(i, j, "%.2f" % nsd, ha="center", va="center", size=12 * 8 // dammif_files)
                    ax1.text(j, i, "%.2f" % nsd, ha="center", va="center", size=12 * 8 // dammif_files)
                    if i != j:
                        lnsd.append(nsd)

        lnsd = numpy.array(lnsd)
        nsd_max = lnsd.mean() + lnsd.std()  # threshold for nsd mean

        ax1.imshow(maskedNSD, interpolation="nearest", origin="upper", cmap="YlOrRd", norm=matplotlib.colors.Normalize(vmin=min(lnsd)))
        ax1.set_title(u"NSD correlation table")
        ax1.set_xticks(range(dammif_files))
        ax1.set_xticklabels(labels, rotation=90)
        ax1.set_xlim(-0.5, dammif_files - 0.5)
        ax1.set_ylim(-0.5, dammif_files - 0.5)
        ax1.set_yticks(range(dammif_files))
        ax1.set_yticklabels(labels)

        # second subplot : the NSD mean for each model
        ax2.bar(xticks - 0.5, data)
        ax2.plot([0.5, dammif_files + 0.5], [nsd_max, nsd_max], "-r", label=u"NSD$_{max}$ = %.2f" % nsd_max)
        ax2.set_title(u"NSD between any model and all others")
        ax2.set_ylabel("Normalized Spatial Discrepancy")
        ax2.set_xticks(xticks)
        ax2.set_xticklabels(labels, rotation=90)
        bbox_props = dict(fc="cyan", ec="b", lw=1)
        ax2.text(self.reference + 0.95, data[self.reference] / 2, "Reference", ha="center", va="center", rotation=90, size=10, bbox=bbox_props)
        ax2.legend(loc=8)

        bbox_props = dict(fc="pink", ec="r", lw=1)
        valid_number = 0
        for i in range(dammif_files):
            if data[i] > nsd_max:
                ax2.text(i + 0.95, data[self.reference] / 2, "Discarded", ha="center", va="center", rotation=90, size=10, bbox=bbox_props)
                logger.debug("model %s discarded, nsd > nsd_max" % self.inputfiles[i])
            elif not valid_models[i]:
                if rmax:
                    ax2.text(i + 0.95, data[self.reference] / 2, "Discarded, Rfactor = %s > Rmax = %s" % (100.0 * self.models[i].rfactor, rmax), ha="center", va="center", rotation=90, size=10, bbox=bbox_props)
                else:
                    ax2.text(i + 0.95, data[self.reference] / 2, "Discarded", ha="center", va="center", rotation=90, size=10, bbox=bbox_props)
            else:
                if valid_models[i] == 1.0:
                    valid_number += 1

        logger.debug("%s valid models" % valid_number)

        if save:
            fig.savefig(filename)
        else:
            fig.show()
        return fig

    def find_reference(self):
        """
        Find the reference model among the models aligned.
        The reference model is the one with lower average NSD with other models.
        
        :return ref_number: position of the reference model in the list self.models
        """
        if self.arrayNSD is None:
            self.makeNSDarray()
        if len(self.validmodels) == 0:
            logger.error("Validity of models is not computed")
        valid = self.validmodels
        valid = valid.astype(bool)

        averNSD = numpy.zeros(len(self.models))
        averNSD += sys.maxsize
        averNSD[valid] = ((self.arrayNSD.sum(axis=-1)) / (valid.sum() - 1))[valid]

        self.reference = averNSD.argmin()

        return self.reference

    def alignment_reference(self, ref_number=None):
        """
        Align all models in self.models with the reference one.
        The aligned models are saved in pdb files (names in list self.outputfiles)
        """
        if self.reference is None and ref_number is None:
            self.find_reference()
        
        ref_number = self.reference
        models = self.models
        reference = models[ref_number]
        for i in range(len(models)):
            if i == ref_number:
                continue
            else:
                molecule = models[i]
                symmetry, p = self.alignment_sym(reference, molecule)
                if not self.slow:
                    p, dist = self.optimize(reference, molecule, symmetry)
                molecule.atoms = molecule.transform(p, symmetry)  # molecule sent on its canonical position
                molecule.atoms = molecule.transform(reference.can_param, [1, 1, 1], reverse=True)  # molecule sent on reference position
                molecule.save(self.outputfiles[i])
        reference.save(self.outputfiles[ref_number])
        return 0

    def alignment_2models(self, save=True):
        """
        Align two models using the first one as reference.
        The aligned models are save in pdb files.

        :return dist: NSD after alignment
        """
        models = self.models
        reference = models[0]
        molecule = models[1]

        symmetry, p = self.alignment_sym(reference, molecule)
        if not self.slow:
            p, dist = self.optimize(reference, molecule, symmetry)

        molecule.atoms = molecule.transform(p, symmetry)
        molecule.atoms = molecule.transform(reference.can_param, [1, 1, 1], reverse=True)
        if self.slow:
            dist = reference.dist(molecule, reference.atoms, molecule.atoms)
        if save:
            molecule.save(self.outputfiles)

        return dist



#OpenMP version of distance calculation

from cython import parallel
cimport openmp 

@cython.wraparound(False)
@cython.boundscheck(False)
def calc_distance(floating[:, :] atoms1, floating[:, :] atoms2, floating fineness1, floating fineness2):
    """
    Calculate the Normalized Spatial Discrepancy (NSD) between two molecules
    
    @param atoms1,atoms2: 2d-array with atom coordinates[[x,y,z],...]
    @param fineness1, fineness2: fineness of each molecule
    @return: NSD atoms1-atoms2
    """
    
    cdef:
        int i, j, size1 = atoms1.shape[0], size2 = atoms2.shape[0]
        int threadid, numthreads = openmp.omp_get_max_threads() 
        double d, d2, dx, dy, dz, x1, y1, z1
        double s1 = 0.0, s2 = 0.0, big = sys.maxsize
        double[:, ::1] min_col = numpy.zeros((numthreads, size2), numpy.float64) + big
    assert atoms1.shape[1] >= 3
    assert atoms2.shape[1] >= 3
    assert size1 > 0
    assert size2 > 0
    
    for i in parallel.prange(size1, nogil=True):
        threadid = parallel.threadid()
        x1 = atoms1[i, 0]
        y1 = atoms1[i, 1]
        z1 = atoms1[i, 2]
        d = big
        for j in range(size2):            
            dx = atoms2[j, 0] - x1
            dy = atoms2[j, 1] - y1
            dz = atoms2[j, 2] - z1
            d2 = dx * dx + dy * dy + dz * dz
            d = min(d, d2)
            min_col[threadid, j] = min(min_col[threadid, j], d2) 
        s1 += d
    for j in parallel.prange(size2, nogil=True):
        d = big
        for i in range(numthreads):
            d = min(d, min_col[i, j])
        s2 += d

    return sqrt(0.5 * ((1.0 / (size1 * fineness2 * fineness2)) * s1 + (1.0 / (size2 * fineness1 * fineness1)) * s2))


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
def calc_density(floating[:, :] atoms, floating dmax, int npt, floating r=0.0, bint hard=True):
    """
    Calculate the density rho(r)
    
    #TODO: formula for rigid sphere:
    A = (4*R+d)*(2*R-d)**2/16.0/R**3
    
    @param atoms: 2d-array with atom coordinates[[x,y,z],...]
    @param dmax: Diameter of the model
    @param npt: number of point in the density
    @param r: radius of an atom
    @param hard: use hard spheres model
    @return: 1d-array of 
    """
    
    cdef:
        int i, j, k, size = atoms.shape[0]        
        int threadid, numthreads = openmp.omp_get_max_threads()
        int width = 1 if hard else 2 
        floating d, dmax_plus, dx, dy, dz, x1, y1, z1
        floating delta, d_min, d_max, d1, den 
        double[:, ::1] tmp = numpy.zeros((numthreads, npt), numpy.float64)
        double[::1] out = numpy.zeros(npt, numpy.float64)
        double s
        
    assert atoms.shape[1] >= 3
    assert size > 0
    assert dmax > 0
    dmax_plus = dmax * (1.0 + numpy.finfo(numpy.float32).eps)
    delta = dmax_plus / npt
    
    for i in parallel.prange(size, nogil=True):
        threadid = parallel.threadid()
        x1 = atoms[i, 0]
        y1 = atoms[i, 1]
        z1 = atoms[i, 2]
        for j in range(size):
            dx = atoms[j, 0] - x1
            dy = atoms[j, 1] - y1
            dz = atoms[j, 2] - z1
            d = sqrt(dx * dx + dy * dy + dz * dz)
            d_min = max(0.0, d - width * r)
            d_max = min(dmax, d + width * r)
            for k in range(<int>(d_min / delta), <int>(d_max / delta)+1):
                if hard:
                    tmp[threadid, k] += hard_sphere(fabs(k * delta - d), r)
                else:
                    tmp[threadid, k] += soft_sphere(fabs(k * delta - d), r)
    for j in parallel.prange(npt, nogil=True):
        s = 0
        for i in range(numthreads):
            s = s + tmp[i, j]
        out[j] += s

    return numpy.asarray(out)



#!/usr/bin/python3
# coding: utf-8
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2020  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"Tool to perform a simple plotting of a set of SAS curve"

__author__ = "Jérôme Kieffer"
__license__ = "MIT"
__copyright__ = "2020, ESRF"
__date__ = "14/05/2020"

import platform
import logging
from pathlib import Path
from matplotlib.pyplot import switch_backend
from matplotlib.backends.backend_pdf import PdfPages
from freesas import plot
from freesas.sasio import (
    load_scattering_data,
    convert_inverse_angstrom_to_nanometer,
)
from freesas.autorg import InsufficientDataError, NoGuinierRegionError
from freesas.sas_argparser import SASParser

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("plot_sas")


def set_backend(output: Path = None, output_format: str = None):
    """Explicitely set silent backend based on format or filename
    Needed on MacOS
    @param output: Name of the specified output file
    @param output_format: User specified format
    """
    if output_format:
        output_format = output_format.lower()
    elif output and len(output.suffix) > 0:
        output_format = output.suffix.lower()[1:]
    if output_format:
        if output_format == "svg":
            switch_backend("svg")
        elif output_format in ["ps", "eps"]:
            switch_backend("ps")
        elif output_format == "pdf":
            switch_backend("pdf")
        elif output_format == "png":
            switch_backend("agg")


def parse():
    """Parse input and return list of files.
    :return: list of input files
    """
    description = "Generate typical sas plots with matplotlib"
    epilog = """freesas is an open-source implementation of a bunch of
    small angle scattering algorithms. """
    parser = SASParser(
        prog="freesas.py", description=description, epilog=epilog
    )
    parser.add_file_argument(help_text="dat files to plot")
    parser.add_output_filename_argument()
    parser.add_output_data_format("jpeg", "svg", "png", "pdf")
    parser.add_q_unit_argument()
    return parser.parse_args()


def create_figure(file: Path, unit: str = "nm"):
    """Create multi-plot SAS figure for data from a file
    @param file: filename of SAS file in q I Ierr format
    @param unit: length unit of input data, supported options are Å and nm.
    :return: figure with SAS plots for this file
    """
    data = load_scattering_data(file)
    if unit == "Å":
        data = convert_inverse_angstrom_to_nanometer(data)
    fig = plot.plot_all(data)
    fig.suptitle(file)
    return fig


def main():
    args = parse()
    if args.verbose:
        logging.root.setLevel(logging.DEBUG)
    files = [Path(i) for i in args.file if Path(i).exists()]
    if platform.system() == "Windows" and files == []:
        files = list(Path.cwd().glob(args.file[0]))
        files.sort()
    input_len = len(files)
    logger.debug("%s input files", input_len)
    figures = []

    if args.output and len(files) > 1:
        logger.warning("Only PDF export is possible in multi-frame mode")
    if args.output and platform.system() == "Darwin":
        if len(files) == 1:
            set_backend(args.output, args.format)
        elif len(files) > 1:
            set_backend(output_format="pdf")
    for afile in files:
        try:
            fig = create_figure(afile, args.unit)
        except OSError:
            logger.error("Unable to load file %s", afile)
        except (InsufficientDataError, NoGuinierRegionError, ValueError):
            logger.error("Unable to process file %s", afile)
        else:
            figures.append(fig)
            if args.output is None:
                fig.show()
            elif len(files) == 1:
                fig.savefig(args.output, format=args.format)
    if len(figures) > 1 and args.output:
        with PdfPages(args.output) as pdf_output_file:
            for fig in figures:
                pdf_output_file.savefig(fig)
    if not args.output:
        input("Press enter to quit")


if __name__ == "__main__":
    main()



__author__ = "Guillaume"
__license__ = "MIT"
__copyright__ = "2015, ESRF"

import numpy
from freesas.model import SASModel


class Grid:
    """
    This class is used to create a grid which include all the input models
    """
    def __init__(self, inputfiles):
        """
        :param inputfiles: list of pdb files needed for averaging
        """
        self.inputs = inputfiles
        self.size = []
        self.nbknots = None
        self.radius = None
        self.coordknots = []

    def __repr__(self):
        return "Grid with %i knots"%self.nbknots

    def spatial_extent(self):
        """
        Calculate the maximal extent of input models
        
        :return self.size: 6-list with x,y,z max and then x,y,z min
        """
        atoms = []
        models_fineness = []
        for files in self.inputs:
            m = SASModel(files)
            if len(atoms)==0:
                atoms = m.atoms
            else:
                atoms = numpy.append(atoms, m.atoms, axis=0)
            models_fineness.append(m.fineness)
        mean_fineness = sum(models_fineness) / len(models_fineness)

        coordmin = atoms.min(axis=0) - mean_fineness
        coordmax = atoms.max(axis=0) + mean_fineness
        self.size = [coordmax[0],coordmax[1],coordmax[2],coordmin[0],coordmin[1],coordmin[2]]

        return self.size

    def calc_radius(self, nbknots=None):
        """
        Calculate the radius of each point of a hexagonal close-packed grid, 
        knowing the total volume and the number of knots in this grid.

        :param nbknots: number of knots wanted for the grid
        :return radius: the radius of each knot of the grid
        """
        if len(self.size)==0:
            self.spatial_extent()
        nbknots = nbknots if nbknots is not None else 5000
        size = self.size
        dx = size[0] - size[3]
        dy = size[1] - size[4]
        dz = size[2] - size[5]
        volume = dx * dy * dz

        density = numpy.pi / (3*2**0.5)
        radius = ((3 /( 4 * numpy.pi)) * density * volume / nbknots)**(1.0/3)
        self.radius = radius

        return radius

    def make_grid(self):
        """
        Create a grid using the maximal size and the radius previously computed.
        The geometry used is a face-centered cubic lattice (fcc).

        :return knots: 2d-array, coordinates of each dot of the grid. Saved as self.coordknots.
        """
        if len(self.size)==0:
            self.spatial_extent()
        if self.radius is None:
            self.calc_radius()

        radius = self.radius
        a = numpy.sqrt(2.0)*radius

        xmax = self.size[0]
        xmin = self.size[3]
        ymax = self.size[1]
        ymin = self.size[4]
        zmax = self.size[2]
        zmin = self.size[5]

        x = 0.0
        y = 0.0
        z = 0.0

        xlist = []
        ylist = []
        zlist = []
        knots = numpy.empty((1,4), dtype="float")
        while (zmin + z) <= zmax:
            zlist.append(z)
            z += a
        while (ymin + y) <= ymax:
            ylist.append(y)
            y += a
        while (xmin + x) <= xmax:
            xlist.append(x)
            x += a

        for i in range(len(zlist)):
            z = zlist[i]
            if i % 2 ==0:
                for j in range(len(xlist)):
                    x = xlist[j]
                    if j % 2 == 0:
                        for y in ylist[0:-1:2]:
                            knots = numpy.append(knots, [[xmin+x, ymin+y, zmin+z, 0.0]], axis=0)
                    else:
                        for y in ylist[1:-1:2]:
                            knots = numpy.append(knots, [[xmin+x, ymin+y, zmin+z, 0.0]], axis=0)
            else:
                for j in range(len(xlist)):
                    x = xlist[j]
                    if j % 2 == 0:
                        for y in ylist[1:-1:2]:
                            knots = numpy.append(knots, [[xmin+x, ymin+y, zmin+z, 0.0]], axis=0)
                    else:
                        for y in ylist[0:-1:2]:
                            knots = numpy.append(knots, [[xmin+x, ymin+y, zmin+z, 0.0]], axis=0)

        knots = numpy.delete(knots, 0, axis=0)
        self.nbknots = knots.shape[0]
        self.coordknots = knots

        return knots


class AverModels():
    """
    Provides tools to create an averaged models using several aligned dummy atom models
    """
    def __init__(self, inputfiles, grid):
        """
        :param inputfiles: list of pdb files of aligned models
        :param grid: 2d-array coordinates of each point of a grid, fourth column full of zeros
        """
        self.inputfiles = inputfiles
        self.models = []
        self.header = []
        self.radius = None
        self.atoms = []
        self.grid = grid

    def __repr__(self):
        return "Average SAS model with %i atoms"%len(self.atoms)

    def read_files(self, reference=None):
        """
        Read all the pdb file in the inputfiles list, creating SASModels.
        The SASModels created are save in a list, the reference model is the first model in the list.

        :param reference: position of the reference model file in the inputfiles list
        """
        ref = reference if reference is not None else 0
        inputfiles = self.inputfiles

        models = []
        models.append(SASModel(inputfiles[ref]))
        for i in range(len(inputfiles)):
            if i==ref:
                continue
            else:
                models.append(SASModel(inputfiles[i]))
        self.models = models

        return models

    def calc_occupancy(self, griddot):
        """
        Assign an occupancy and a contribution factor to the point of the grid.

        :param griddot: 1d-array, coordinates of a point of the grid
        :return tuple: 2-tuple containing (occupancy, contribution)
        """
        occ = 0.0
        contrib = 0
        for model in self.models:
            f = model.fineness
            for i in range(model.atoms.shape[0]):
                dx = model.atoms[i, 0] - griddot[0]
                dy = model.atoms[i, 1] - griddot[1]
                dz = model.atoms[i, 2] - griddot[2]
                dist = dx * dx + dy * dy + dz * dz
                add = max(1 - (dist / f), 0)
                if add != 0:
                    contrib += 1
                    occ += add
        return occ, contrib

    def assign_occupancy(self):
        """
        For each point of the grid, total occupancy and contribution factor are computed and saved.
        The grid is then ordered with decreasing value of occupancy.
        The fourth column of the array correspond to the occupancy of the point and the fifth to 
        the contribution for this point.

        :return sortedgrid: 2d-array, coordinates of each point of the grid
        """
        grid = self.grid
        nbknots = grid.shape[0]
        grid = numpy.append(grid, numpy.zeros((nbknots, 1), dtype="float"), axis=1)

        for i in range(nbknots):
            occ, contrib = self.calc_occupancy(grid[i, 0:3])
            grid[i, 3] = occ
            grid[i, 4] = contrib

        order = numpy.argsort(grid, axis=0)[:, -2]
        sortedgrid = numpy.empty_like(grid)
        for i in range(nbknots):
            sortedgrid[nbknots - i - 1, :] = grid[order[i], :]

        return sortedgrid

    def make_header(self):
        """
        Create the layout of the pdb file for the averaged model.
        """
        header = []
        header.append("Number of files averaged : %s\n"%len(self.inputfiles))
        for i in self.inputfiles:
            header.append(i + "\n")
        header.append("Total number of dots in the grid : %s\n"%self.grid.shape[0])

        decade = 1
        for i in range(self.grid.shape[0]):
            line = "ATOM         CA  ASP    1                                    20.00   2 201\n"
            line = line[:7] + "%4.i"%(i + 1) + line[11:]
            if not (i + 1) % 10:
                decade += 1
            line = line[:21] + "%4.i"%decade + line[25:]
            header.append(line)
        self.header = header
        return header

    def save_aver(self, filename):
        """
        Save the position of each occupied dot of the grid, its occupancy and its contribution 
        in a pdb file.

        :param filename: name of the pdb file to write
        """
        if len(self.header) == 0:
            self.make_header()
        assert self.grid.shape[-1] == 5

        nr = 0
        with open(filename, "w") as pdbout:
            for line in self.header:
                if line.startswith("ATOM"):
                    if nr < self.grid.shape[0] and self.grid[nr, 4] != 0:
                        coord = "%8.3f%8.3f%8.3f" % tuple(self.grid[nr, 0:3])
                        occ = "%6.2f" % self.grid[nr, 3]
                        contrib = "%2.f" % self.grid[nr, 4]
                        line = line[:30] + coord + occ + line[60:66] + contrib + line[68:]
                    else:
                        line = ""
                    nr += 1
                pdbout.write(line)



# -*- coding: utf-8 -*-
"""
Bayesian Inverse Fourier Transform

This code is the implementation of
Steen Hansen J. Appl. Cryst. (2000). 33, 1415-1421

Based on the BIFT from Jesse Hopkins, available at:
https://sourceforge.net/p/bioxtasraw/git/ci/master/tree/bioxtasraw/BIFT.py

Many thanks to Pierre Paleo for the auto-alpha guess
"""

__authors__ = ["Jerome Kieffer", "Jesse Hopkins"]
__license__ = "MIT"
__copyright__ = "2020, ESRF"
__date__ = "10/06/2020"

import logging
logger = logging.getLogger(__name__)
# from collections import namedtuple
from math import log, ceil
import numpy
from scipy.optimize import minimize
from ._bift import BIFT
from .autorg import auto_gpa, autoRg, auto_guinier, NoGuinierRegionError


def auto_bift(data, Dmax=None, alpha=None, npt=100,
              start_point=None, end_point=None,
              scan_size=11, Dmax_over_Rg=3):
    """Calculates the inverse Fourier tranform of the data using an optimisation of the evidence

    :param data: 2D array with q, I(q), δI(q). q can be in 1/nm or 1/A, it imposes the unit for r & Dmax
    :param Dmax: Maximum diameter of the object, this is the starting point to be refined. Can be guessed
    :param alpha: Regularisation parameter, let it to None for automatic scan
    :param npt: Number of point for the curve p(r)
    :param start_point: First useable point in the I(q) curve, this is not the start of the Guinier region
    :param end_point: Last useable point in the I(q) curve
    :param scan_size: size of the initial geometrical scan for alpha values.
    :param Dmax_over_Rg: In average, protein's Dmax is 3x Rg, use this to adjust
    :return: BIFT object. Call the get_best to retrieve the optimal solution
    """
    assert data.ndim == 2
    assert data.shape[1] == 3  # enforce q, I, err
    use_wisdom = False
    data = data[slice(start_point, end_point)]
    q, I, err = data.T
    npt = min(npt, q.size)  # no chance for oversampling !
    bo = BIFT(q, I, err)  # this is the bift object
    if Dmax is None:
        # Try to get a reasonable guess from Rg
        try:
            Guinier = auto_guinier(data)
        except:
            logger.error("Guinier analysis failed !")
            raise
#         print(Guinier)
        if Guinier.Rg <= 0:
            raise NoGuinierRegionError
        Dmax = bo.set_Guinier(Guinier, Dmax_over_Rg)
    if alpha is None:
        alpha_max = bo.guess_alpha_max(npt)
        # First scan on alpha:
        key = bo.grid_scan(Dmax, Dmax, 1,
                           1.0 / alpha_max, alpha_max, scan_size, npt)
        Dmax, alpha = key[:2]
        # Then scan on Dmax:
        key = bo.grid_scan(max(Dmax / 2, Dmax * (Dmax_over_Rg - 1) / Dmax_over_Rg), Dmax * (Dmax_over_Rg + 1) / Dmax_over_Rg, scan_size,
                           alpha, alpha, 1, npt)
        Dmax, alpha = key[:2]
        if bo.evidence_cache[key].converged:
            bo.update_wisdom()
            use_wisdom = True

    # Optimization using Bayesian operator:
    logger.info("Start search at Dmax=%.2f alpha=%.2f use wisdom=%s", Dmax, alpha, use_wisdom)
    res = minimize(bo.opti_evidence, (Dmax, log(alpha)), args=(npt, use_wisdom), method="powell")
    logger.info("Result of optimisation:\n  %s", res)
    return bo



# -*- coding: utf-8 -*-
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2017  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__authors__ = ["J. Kieffer"]
__license__ = "MIT"
__date__ = "10/06/2020"

import numpy
import unittest
from .utilstests import get_datafile
from ..bift import auto_bift
from .._bift import BIFT, distribution_parabola, distribution_sphere, \
                    ensure_edges_zero, smooth_density
import logging
logger = logging.getLogger(__name__)
import time


class TestBIFT(unittest.TestCase):

    DMAX = 10
    NPT = 100
    SIZE = 1000

    @classmethod
    def setUpClass(cls):
        super(TestBIFT, cls).setUpClass()
        cls.r = numpy.linspace(0, cls.DMAX, cls.NPT + 1)
        dr = cls.DMAX / cls.NPT
        cls.p = -cls.r * (cls.r - cls.DMAX)  # Nice parabola
        q = numpy.linspace(0, 8 * cls.DMAX / 3, cls.SIZE + 1)
        sincqr = numpy.sinc(numpy.outer(q, cls.r / numpy.pi))
        I = 4 * numpy.pi * (cls.p * sincqr).sum(axis=-1) * dr
        err = numpy.sqrt(I)
        cls.I0 = I[0]
        cls.q = q[1:]
        cls.I = I[1:]
        cls.err = err[1:]
        cls.Rg = numpy.sqrt(0.5 * numpy.trapz(cls.p * cls.r ** 2, cls.r) / numpy.trapz(cls.p, cls.r))
        print(cls.Rg)

    @classmethod
    def tearDownClass(cls):
        super(TestBIFT, cls).tearDownClass()
        cls.r = cls.p = cls.I = cls.q = cls.err = None

    def test_autobift(self):
        data = numpy.vstack((self.q, self.I, self.err)).T
        t0 = time.perf_counter()
        bo = auto_bift(data)
        key, value, valid = bo.get_best()
#         print("key is ", key)
        stats = bo.calc_stats()
#         print("stat is ", stats)
        logger.info("Auto_bift time: %s", time.perf_counter() - t0)
        self.assertAlmostEqual(self.DMAX / key.Dmax, 1, 1, "DMax is correct")
        self.assertAlmostEqual(self.I0 / stats.I0_avg, 1, 1, "I0 is correct")
        self.assertAlmostEqual(self.Rg / stats.Rg_avg, 1, 2, "Rg is correct")

    def test_BIFT(self):
        bift = BIFT(self.q, self.I, self.err)
        # test two scan functions
        key = bift.grid_scan(9, 11, 5, 10, 100, 5, 100)
        # print("key is ", key)
        self.assertAlmostEqual(self.DMAX / key.Dmax, 1, 2, "DMax is correct")
        res = bift.monte_carlo_sampling(10, 3, 100)
        # print("res is ", res)
        self.assertAlmostEqual(self.DMAX / res.Dmax_avg, 1, 4, "DMax is correct")

    def test_disributions(self):
        pp = numpy.asarray(distribution_parabola(self.I0, self.DMAX, self.NPT))
        ps = numpy.asarray(distribution_sphere(self.I0, self.DMAX, self.NPT))
        self.assertAlmostEqual(numpy.trapz(ps, self.r) * 4 * numpy.pi / self.I0, 1, 3, "Distribution for a sphere looks OK")
        self.assertAlmostEqual(numpy.trapz(pp, self.r) * 4 * numpy.pi / self.I0, 1, 3, "Distribution for a parabola looks OK")
        self.assertTrue(numpy.allclose(pp, self.p, 1e-4), "distribution matches")

    def test_fixEdges(self):
        ones = numpy.ones(self.NPT)
        ensure_edges_zero(ones)
        self.assertAlmostEqual(ones[0], 0,  msg="1st point set to 0")
        self.assertAlmostEqual(ones[-1], 0,  msg="last point set to 0")
        self.assertTrue(numpy.allclose(ones[1:-1], numpy.ones(self.NPT-2), 1e-7), msg="non-edge points unchanged")

    def test_smoothing(self):
        ones = numpy.ones(self.NPT)
        empty = numpy.empty(self.NPT)
        smooth_density(ones,empty)
        self.assertTrue(numpy.allclose(ones, empty, 1e-7), msg="flat array smoothed into flat array")
        random = numpy.random.rand(self.NPT)
        smooth =  numpy.empty(self.NPT)
        smooth_density(random,smooth)
        self.assertAlmostEqual(random[0], smooth[0],  msg="first points of random array and smoothed random array match")
        self.assertAlmostEqual(random[-1], smooth[-1],  msg="last points of random array and smoothed random array match")
        self.assertTrue(smooth[1]>=min(smooth[0], smooth[2]) and smooth[1]<=max(smooth[0], smooth[2]), msg="second point of random smoothed array between 1st and 3rd")
        self.assertTrue(smooth[-2]>=min(smooth[-1], smooth[-3]) and smooth[-2]<= max(smooth[-1], smooth[-3]), msg="second to last point of random smoothed array between 3rd to last and last")
        sign = numpy.sign(random[1:-3] - smooth[2:-2]) * numpy.sign(smooth[2:-2] - random[3:-1])
        self.assertTrue(numpy.allclose(sign, numpy.ones(self.NPT-4), 1e-7), msg="central points of random array and smoothed random array alternate")

def suite():
    testSuite = unittest.TestSuite()
    testSuite.addTest(TestBIFT("test_disributions"))
    testSuite.addTest(TestBIFT("test_autobift"))
    testSuite.addTest(TestBIFT("test_fixEdges"))
    testSuite.addTest(TestBIFT("test_smoothing"))
    return testSuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/python
# coding: utf-8

"""Test the functionality of SASParser and GuinierParser"""

__authors__ = ["Martha Brennich"]
__license__ = "MIT"
__date__ = "25/03/2021"


import unittest
import logging
import io
import contextlib
from pathlib import Path
from .. import dated_version as freesas_version
from ..sas_argparser import SASParser, GuinierParser


logger = logging.getLogger(__name__)


class TestSasArgParser(unittest.TestCase):
    def test_minimal_guinier_parser_requires_file_argument(self):
        """
        Test that Guinier parser provides error if no file argument is provided.
        """
        basic_parser = GuinierParser("program", "description", "epilog")
        output_catcher = io.StringIO()
        try:
            with contextlib.redirect_stderr(output_catcher):
                _ = basic_parser.parse_args()
        except SystemExit:
            pass

        self.assertTrue(
            basic_parser.usage in output_catcher.getvalue(),
            msg="GuinierParser provides usage if no file provided",
        )
        self.assertTrue(
            "the following arguments are required: FILE"
            in output_catcher.getvalue(),
            msg="GuinierParser states that the FILE argument is missing if no file provided",
        )

    def test_minimal_guinier_parser_parses_list_of_files(self):
        """
        Test that the Guinier parsers parses a list of files.
        """
        basic_parser = GuinierParser("program", "description", "epilog")

        parsed_arguments = basic_parser.parse_args(["afile", "bfile", "cfile"])

        self.assertEqual(
            set(parsed_arguments.file),
            {"afile", "bfile", "cfile"},
            msg="GuinierParser parses list of files",
        )

    def test_add_file_argument_enables_SASParser_to_recognize_file_lists(
        self,
    ):
        """
        Test that add_file_argument adds the ability to parse a file list to SASParser.
        """
        basic_parser = SASParser("program", "description", "epilog")

        # Before running add_file_argument a file argument is not recognized
        output_catcher = io.StringIO()
        try:
            with contextlib.redirect_stderr(output_catcher):
                _ = basic_parser.parse_args(["afile"])
        except SystemExit:
            pass
        self.assertTrue(
            "unrecognized arguments: afile" in output_catcher.getvalue(),
            msg="Minimal SASParser does not recognize file argument",
        )

        basic_parser.add_file_argument(help_text="file help")
        parsed_arguments = basic_parser.parse_args(["afile", "bfile", "cfile"])

        self.assertEqual(
            set(parsed_arguments.file),
            {"afile", "bfile", "cfile"},
            msg="GuinierParser parses list of files",
        )

    def test_minimal_parser_usage_includes_program_name(self):
        """
        Test that minimal parser includes the provided program in the usage string.
        """
        basic_parser = SASParser("test❤️", "description", "epilog")

        self.assertTrue(
            "test❤️" in basic_parser.usage,
            msg="SASParser usage includes program name",
        )

    def test_minimal_guinier_parser_usage_includes_program_name(self):
        """
        Test that minimal parser includes the provided program in the usage string.
        """
        basic_parser = GuinierParser("test❤️", "description", "epilog")

        self.assertTrue(
            "test❤️" in basic_parser.usage,
            msg="GuinierParser usage includes program name",
        )

    def test_minimal_guinier_parser_help_includes_program_description_epilog(
        self,
    ):
        """
        Test that minimal guinier parser includes help includes
        the provided program name, description and epilog.
        """
        basic_parser = GuinierParser("test❤️", "description📚", "epilog🎦")
        output_catcher = io.StringIO()

        try:
            with contextlib.redirect_stdout(output_catcher):
                _ = basic_parser.parse_args(["--help"])
        except SystemExit:
            pass

        self.assertTrue(
            "test❤️" in output_catcher.getvalue(),
            msg="GuinierParser outputs program name in help",
        )

        self.assertTrue(
            "description📚" in output_catcher.getvalue(),
            msg="GuinierParser outputs description in help",
        )

        self.assertTrue(
            "epilog🎦" in output_catcher.getvalue(),
            msg="GuinierParser outputs eplilog name in help",
        )

    def test_minimal_parser_help_includes_program_description_epilog(self):
        """
        Test that minimal parser includes help includes
        the provided program name, description and epilog.
        """
        basic_parser = SASParser("test❤️", "description📚", "epilog🎦")
        output_catcher = io.StringIO()

        try:
            with contextlib.redirect_stdout(output_catcher):
                _ = basic_parser.parse_args(["--help"])
        except SystemExit:
            pass

        self.assertTrue(
            "test❤️" in output_catcher.getvalue(),
            msg="SASParser outputs program name in help",
        )

        self.assertTrue(
            "description📚" in output_catcher.getvalue(),
            msg="SASParser outputs description in help",
        )

        self.assertTrue(
            "epilog🎦" in output_catcher.getvalue(),
            msg="SASParser outputs eplilog name in help",
        )

    def test_minimal_parser_default_verbosity_level_is_0(self):
        """
        Test that the parser sets the verbosity to 0 if no args are provided
        """
        basic_parser = SASParser("program", "description", "epilog")
        parsed_arguments = basic_parser.parse_args()
        self.assertEqual(
            parsed_arguments.verbose,
            0,
            msg="SASParser default verbosity is 0",
        )

    def test_minimal_guinier_parser_default_verbosity_level_is_0(self):
        """
        Test that the Guinier parser sets the verbosity to 0 if no args are provided
        """
        basic_parser = GuinierParser("program", "description", "epilog")
        parsed_arguments = basic_parser.parse_args(["afile"])
        self.assertEqual(
            parsed_arguments.verbose,
            0,
            msg="GuinierParser default verbosity is 0",
        )

    def test_minimal_parser_accumulates_verbosity_level(self):
        """
        Test that the parser parser increases the verbosity level to two
        if -vv argument is provided.
        """
        basic_parser = SASParser("program", "description", "epilog")
        parsed_arguments = basic_parser.parse_args(["-vv"])
        self.assertEqual(
            parsed_arguments.verbose,
            2,
            msg="SASParser verbosity increased to 2 by -vv",
        )

    def test_minimal_guinier_parser_accumulates_verbosity_level(self):
        """
        Test that the parser parser increases the verbosity level to two
        if -vv argument is provided.
        """
        basic_parser = GuinierParser("program", "description", "epilog")
        parsed_arguments = basic_parser.parse_args(["afile", "-vv"])
        self.assertEqual(
            parsed_arguments.verbose,
            2,
            msg="GuinierParser verbosity increased to 2 by -vv",
        )

    def test_minimal_parser_provides_correct_version(self):
        """
        Test that parser provides the correct app version.
        """
        basic_parser = SASParser("program", "description", "epilog")
        output_catcher = io.StringIO()
        try:
            with contextlib.redirect_stdout(output_catcher):
                _ = basic_parser.parse_args(["--version"])
        except SystemExit:
            pass

        self.assertTrue(
            freesas_version.version in output_catcher.getvalue(),
            msg="SASParser outputs consistent version",
        )
        self.assertTrue(
            freesas_version.date in output_catcher.getvalue(),
            msg="SASParser outputs consistent date",
        )

    def test_minimal_guinier_parser_provides_correct_version(self):
        """
        Test that parser provides the correct app version.
        """
        basic_parser = GuinierParser("program", "description", "epilog")
        output_catcher = io.StringIO()
        try:
            with contextlib.redirect_stdout(output_catcher):
                _ = basic_parser.parse_args(["--version"])
        except SystemExit:
            pass

        self.assertTrue(
            freesas_version.version in output_catcher.getvalue(),
            msg="GuinierParser outputs consistent version",
        )
        self.assertTrue(
            freesas_version.date in output_catcher.getvalue(),
            msg="GuinierParser outputs consistent date",
        )

    def test_minimal_guinier_parser_accepts_output_file_argument(self):
        """
        Test that minimal Guinier parser accepts one output file argument.
        """
        basic_parser = GuinierParser("program", "description", "epilog")
        parsed_arguments = basic_parser.parse_args(["afile", "-o", "out.file"])

        self.assertEqual(
            parsed_arguments.output,
            Path("out.file"),
            msg="Minimal GuinierParser accepts output file argument",
        )

    def test_add_output_filename_argument_adds_output_file_argument_to_SASParser(
        self,
    ):
        """
        Test that add_output_filename_argument adds one output file argument to as SASParser.
        """
        basic_parser = SASParser("program", "description", "epilog")

        # Before running add_output_filename_argument -o file is not regognized
        output_catcher = io.StringIO()
        try:
            with contextlib.redirect_stderr(output_catcher):
                _ = basic_parser.parse_args(["-o", "out.file"])
        except SystemExit:
            pass
        self.assertTrue(
            "unrecognized arguments: -o out.file" in output_catcher.getvalue(),
            msg="Minimal SASParser does not recognize -o argument",
        )

        basic_parser.add_output_filename_argument()
        parsed_arguments = basic_parser.parse_args(["-o", "out.file"])

        self.assertEqual(
            parsed_arguments.output,
            Path("out.file"),
            msg="SASParser accepts output file argument"
            "after running add_output_filename_argument()",
        )

    def test_minimal_guinier_parser_accepts_output_format_argument(self):
        """
        Test that minimal Guinier parser accepts one output data format argument.
        """
        basic_parser = GuinierParser("program", "description", "epilog")
        parsed_arguments = basic_parser.parse_args(["afile", "-f", "aformat"])

        self.assertEqual(
            parsed_arguments.format,
            "aformat",
            msg="Minimal GuinierParser accepts output data format argument",
        )

    def test_add_output_data_format_adds_output_format_argument_to_SASParser(
        self,
    ):
        """
        Test that add_output_data_format adds one output data format argument to as SASParser.
        """
        basic_parser = SASParser("program", "description", "epilog")

        # Before running add_output_filename_argument -o file is not regognized
        output_catcher = io.StringIO()
        try:
            with contextlib.redirect_stderr(output_catcher):
                _ = basic_parser.parse_args(["-f", "aformat"])
        except SystemExit:
            pass
        self.assertTrue(
            "unrecognized arguments: -f aformat" in output_catcher.getvalue(),
            msg="Minimal SASParser does not recognize -f argument",
        )

        basic_parser.add_output_data_format()
        parsed_arguments = basic_parser.parse_args(["-f", "aformat"])

        self.assertEqual(
            parsed_arguments.format,
            "aformat",
            msg="SASParser accepts output data format argument"
            "after running add_output_data_format()",
        )

    def test_minimal_guinier_parser_accepts_q_unit_argument(self):
        """
        Test that minimal Guinier parser accepts a q unit argument.
        """
        basic_parser = GuinierParser("program", "description", "epilog")
        parsed_arguments = basic_parser.parse_args(["afile", "-u", "nm"])

        self.assertEqual(
            parsed_arguments.unit,
            "nm",
            msg="Minimal GuinierParser accepts q unit argument",
        )

    def test_add_q_unit_argument_adds_add_q_unit_argument_to_SASParser(
        self,
    ):
        """
        Test that add_q_unit_argument adds a q unit argument to as SASParser.
        """
        basic_parser = SASParser("program", "description", "epilog")

        # Before running add_output_filename_argument -o file is not regognized
        output_catcher = io.StringIO()
        try:
            with contextlib.redirect_stderr(output_catcher):
                _ = basic_parser.parse_args(["-u", "nm"])
        except SystemExit:
            pass
        self.assertTrue(
            "unrecognized arguments: -u nm" in output_catcher.getvalue(),
            msg="Minimal SASParser does not recognize -u argument",
        )

        basic_parser.add_q_unit_argument()
        parsed_arguments = basic_parser.parse_args(["-u", "nm"])

        self.assertEqual(
            parsed_arguments.unit,
            "nm",
            msg="SASParser accepts q unit argument after running add_q_unit_argument()",
        )

    def test_SASParser_q_unit_argument_allows_predefined_units(
        self,
    ):
        """
        Test that the q unit argument of a SASparser accepts "nm", "Å", "A".
        """
        basic_parser = SASParser("program", "description", "epilog")
        basic_parser.add_q_unit_argument()

        parsed_arguments = basic_parser.parse_args(["-u", "nm"])
        self.assertEqual(
            parsed_arguments.unit,
            "nm",
            msg="SASParser accepts unit format nm",
        )

        parsed_arguments = basic_parser.parse_args(["-u", "A"])
        self.assertEqual(
            parsed_arguments.unit,
            "Å",
            msg="SASParser accepts unit format A",
        )

        parsed_arguments = basic_parser.parse_args(["-u", "Å"])
        self.assertEqual(
            parsed_arguments.unit,
            "Å",
            msg="SASParser accepts unit format A",
        )

    def test_SASParser_q_unit_argument_does_not_allow_not_predefined_units(
        self,
    ):
        """
        Test that the q unit argument of a SASparser does not accept a
        unit that is not "nm", "Å", "A".
        """
        basic_parser = SASParser("program", "description", "epilog")
        basic_parser.add_q_unit_argument()

        output_catcher = io.StringIO()
        try:
            with contextlib.redirect_stderr(output_catcher):
                _ = basic_parser.parse_args(["-u", "m"])
        except SystemExit:
            pass
        self.assertTrue(
            "argument -u/--unit: invalid choice: 'm' (choose from 'nm', 'Å', 'A')"
            in output_catcher.getvalue(),
            msg="SASParser does not accept '-u m' argument",
        )

    def test_SASParser_q_unit_A_gets_converted_to_Å(
        self,
    ):
        """
        Test that for a SASParder q unit input "A" gets converted to "Å".
        """
        basic_parser = SASParser("program", "description", "epilog")
        basic_parser.add_q_unit_argument()

        parsed_arguments = basic_parser.parse_args(["-u", "A"])
        self.assertEqual(
            parsed_arguments.unit,
            "Å",
            msg="SASParser converts unit input 'A' to 'Å'",
        )

    def test_GuinierParser_q_unit_argument_allows_predefined_units(
        self,
    ):
        """
        Test that the q unit argument of a Guinierparser accepts "nm", "Å", "A".
        """
        basic_parser = GuinierParser("program", "description", "epilog")

        parsed_arguments = basic_parser.parse_args(["afile", "-u", "nm"])
        self.assertEqual(
            parsed_arguments.unit,
            "nm",
            msg="SASParser accepts unit format nm",
        )

        parsed_arguments = basic_parser.parse_args(["afile", "-u", "A"])
        self.assertEqual(
            parsed_arguments.unit,
            "Å",
            msg="SASParser accepts unit format A",
        )

        parsed_arguments = basic_parser.parse_args(["afile", "-u", "Å"])
        self.assertEqual(
            parsed_arguments.unit,
            "Å",
            msg="SASParser accepts unit format A",
        )

    def test_GuinierParser_q_unit_argument_does_not_allow_not_predefined_units(
        self,
    ):
        """
        Test that the q unit argument of a Guinierparser does not accept a
        unit that is not "nm", "Å", "A".
        """
        basic_parser = GuinierParser("program", "description", "epilog")

        output_catcher = io.StringIO()
        try:
            with contextlib.redirect_stderr(output_catcher):
                _ = basic_parser.parse_args(["afile", "-u", "m"])
        except SystemExit:
            pass
        self.assertTrue(
            "argument -u/--unit: invalid choice: 'm' (choose from 'nm', 'Å', 'A')"
            in output_catcher.getvalue(),
            msg="SASParser does not accept '-u m' argument",
        )

    def test_GuinierParser_q_unit_A_gets_converted_to_Å(
        self,
    ):
        """
        Test that for a GuinierParser q unit input "A" gets converted to "Å".
        """
        basic_parser = GuinierParser("program", "description", "epilog")

        parsed_arguments = basic_parser.parse_args(["afile", "-u", "A"])
        self.assertEqual(
            parsed_arguments.unit,
            "Å",
            msg="SASParser converts unit input 'A' to 'Å'",
        )

    def test_add_argument_adds_an_argument_to_a_SASParser(
        self,
    ):
        """
        Test that new arguments can be added to SASParser.
        """
        basic_parser = SASParser("program", "description", "epilog")

        # Before running add_argument -c
        output_catcher = io.StringIO()
        try:
            with contextlib.redirect_stderr(output_catcher):
                _ = basic_parser.parse_args(["-c"])
        except SystemExit:
            pass
        self.assertTrue(
            "unrecognized arguments: -c" in output_catcher.getvalue(),
            msg="Minimal SASParser does not recognize -c argument",
        )

        basic_parser.add_argument(
            "-c",
            "--check",
            action="store_true",
        )

        parsed_arguments = basic_parser.parse_args(["-c"])
        self.assertEqual(
            parsed_arguments.check,
            True,
            msg="-c argument added to SASParser",
        )

    def test_add_argument_adds_an_argument_to_a_GuinierParser(
        self,
    ):
        """
        Test that new arguments can be added to GuinierParser.
        """
        basic_parser = GuinierParser("program", "description", "epilog")

        # Before running add_argument -c
        output_catcher = io.StringIO()
        try:
            with contextlib.redirect_stderr(output_catcher):
                _ = basic_parser.parse_args(["afile", "-c"])
        except SystemExit:
            pass
        print(output_catcher.getvalue())
        self.assertTrue(
            "unrecognized arguments: -c" in output_catcher.getvalue(),
            msg="Minimal GuinierParser does not recognize -c argument",
        )

        basic_parser.add_argument(
            "-c",
            "--check",
            action="store_true",
        )

        parsed_arguments = basic_parser.parse_args(["afile", "-c"])
        self.assertEqual(
            parsed_arguments.check,
            True,
            msg="-c argument added to GuinierParser",
        )


def suite():
    """Build a test suite from the TestSasArgParser class"""
    test_suite = unittest.TestSuite()
    for class_element in dir(TestSasArgParser):
        if class_element.startswith("test"):
            test_suite.addTest(TestSasArgParser(class_element))
    return test_suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python
# coding: utf-8

"""Run the end to end tests of the project."""

__author__ = "Martha Brennich"
__license__ = "MIT"
__copyright__ = "2020"
__date__ = "11/07/2020"

import sys
import unittest
import e2etest_freesas, e2etest_guinier_apps, e2etest_bift


def suite():
    """Creates suite for e2e tests"""
    test_suite = unittest.TestSuite()
    test_suite.addTest(e2etest_freesas.suite())
    test_suite.addTest(e2etest_guinier_apps.suite())
    test_suite.addTest(e2etest_bift.suite())
    return test_suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    result = runner.run(suite())

    if result.wasSuccessful():
        EXIT_STATUS = 0
    else:
        EXIT_STATUS = 1

    sys.exit(EXIT_STATUS)



#!/usr/bin/python
# coding: utf-8

"""Test the functionality of fitting module."""

__authors__ = ["Martha Brennich"]
__license__ = "MIT"
__date__ = "16/07/2021"


import unittest
from unittest.mock import patch, MagicMock
import logging
import sys
import importlib
import platform
from io import StringIO
import pathlib
import contextlib
from types import SimpleNamespace
from typing import Callable
from errno import ENOENT
import numpy
from ..fitting import (
    set_logging_level,
    get_output_destination,
    get_header,
    rg_result_to_output_line,
    get_linesep,
    run_guinier_fit,
)
from ..autorg import RG_RESULT, InsufficientDataError, NoGuinierRegionError
from ..sas_argparser import GuinierParser

if sys.version_info.minor > 6:
    from unittest.mock import mock_open
else:
    from .mock_open_38 import mock_open


logger = logging.getLogger(__name__)


def reload_os_and_fitting():
    """Some tests patch os and need to reload the modules."""
    os = importlib.import_module("os")
    os = importlib.reload(os)
    fit = importlib.import_module("..fitting", "freesas.subpkg")
    fit = importlib.reload(fit)
    return fit


def get_dummy_guinier_parser(**parse_output):
    """Function which provides a fake GuinierParser with a predefined parse result."""

    def get_mock_parse(**kwargs):
        def mock_parse():
            return SimpleNamespace(**kwargs)

        return mock_parse

    parser = GuinierParser(prog="test", description="test", epilog="test")
    parser.parse_args = get_mock_parse(**parse_output)
    return parser


def patch_linesep(test_function):
    """Patch fitting.linesep to "linesep"."""

    linesep_patch = patch(
        "freesas.fitting.get_linesep",
        MagicMock(return_value="linesep"),
    )
    return linesep_patch(test_function)


def patch_collect_files(test_function):
    """Patch fitting.collect_files to return Paths "test" and "test2"."""

    collect_files_patch = patch(
        "freesas.fitting.collect_files",
        MagicMock(return_value=[pathlib.Path("test"), pathlib.Path("test2")]),
    )
    return collect_files_patch(test_function)


def counted(function: Callable) -> Callable:
    """Wrapper for functions to keep track on how often it has been called."""

    def wrapped(*args, **kwargs):
        wrapped.calls += 1
        return function(*args, **kwargs)

    wrapped.calls = 0
    return wrapped


def build_mock_for_load_scattering_with_Errors(erronous_file: dict):

    """Create mock for loading of data from a file.
    The resulting function will raise an error,
    for files for which an error is provided in errenous_file,
    and an ndarry for all other files."""

    def mock_for_load_scattering(file: pathlib.Path):
        if file.name in erronous_file:
            raise erronous_file[file.name]
        else:
            return numpy.array(
                [[1.0, 1.0, 1.0], [2.0, 2.0, 1.0], [3.0, 3.0, 3.0]]
            )

    return mock_for_load_scattering


class TestFitting(unittest.TestCase):
    def test_set_logging_level_does_not_change_logging_level_if_input_lower_1(
        self,
    ):

        """
        Test that the logging level only gets changed if the requested level is > 0.
        """

        initial_logging_level = logging.root.level
        set_logging_level(0)
        self.assertEqual(
            logging.root.level,
            initial_logging_level,
            msg="setting verbosity to 0 dos not affect logging level",
        )
        set_logging_level(-2)
        self.assertEqual(
            logging.root.level,
            initial_logging_level,
            msg="settting verbosity to -2 does not affect logging level",
        )
        # Ensure that initial logging level is restored
        logging.root.setLevel(initial_logging_level)

    def test_set_logging_level_sets_logging_to_INFO_if_input_is_1(
        self,
    ):

        """
        Test that the logging level gets changed to INFO if verbosity is 1.
        """
        initial_logging_level = logging.root.level
        # Ensure that the function actually changes the level
        logging.root.setLevel(logging.WARNING)

        set_logging_level(1)
        self.assertEqual(
            logging.root.level,
            logging.INFO,
            msg="settting verbosity to 1 sets logging level to INFO",
        )

        # Ensure that initial logging level is restored
        logging.root.setLevel(initial_logging_level)

    def test_set_logging_level_sets_logging_to_DEBUG_if_input_is_2_or_more(
        self,
    ):

        """
        Test that the logging level gets changed to DEBUG if verbosity is 2 or larger.
        """
        initial_logging_level = logging.root.level
        # Ensure that the function actually changes the level
        logging.root.setLevel(logging.WARNING)

        set_logging_level(2)
        self.assertEqual(
            logging.root.level,
            logging.DEBUG,
            msg="settting verbosity to 2 sets logging level to DEBUG",
        )

        set_logging_level(3)
        self.assertEqual(
            logging.root.level,
            logging.DEBUG,
            msg="settting verbosity to 3 sets logging level to DEBUG",
        )

        # Ensure that initial logging level is restored
        logging.root.setLevel(initial_logging_level)

    @patch.dict("sys.modules", {"nt": MagicMock()})
    def test_get_linesep_returns_rn_if_output_is_stdout_on_windows(self):

        """
        Test that get_linesep() returns \r\n if output destination is sys.stdout on Windows.
        """
        # Reload to apply patches
        with patch("sys.builtin_module_names", ["nt"]):
            fit = reload_os_and_fitting()

        self.assertEqual(fit.get_linesep(sys.stdout), "\r\n")

        # Cleanup
        reload_os_and_fitting()

    @unittest.skipIf(platform.system() == "Windows", "Only POSIX")
    def test_get_linesep_returns_n_if_output_is_stdout_on_posix(
        self,
    ):

        """
        Test that get_linesep() returns \n if output destination is sys.stdout on Posix.
        Only should run on posix.
        """
        self.assertEqual(get_linesep(sys.stdout), "\n")

    @patch.dict("sys.modules", {"nt": MagicMock()})
    def test_get_linesep_returns_n_if_output_is_filestream_on_windows(self):

        """
        Test that get_linesep() returns \n if output destination is a filestream on Windows.
        """
        # Reload to apply patches
        with patch("sys.builtin_module_names", ["nt"]):
            fit = reload_os_and_fitting()
        output_dest = StringIO()
        self.assertEqual(fit.get_linesep(output_dest), "\n")

        # Cleanup
        _ = reload_os_and_fitting()

    @unittest.skipIf(platform.system() == "Windows", "Only POSIX")
    def test_get_linesep_returns_n_if_output_is_filestream_on_posix(
        self,
    ):

        """
        Test that get_linesep() returns \n if output destination is filestream on Posix.
        Only should run on posix.
        """
        output_dest = StringIO()
        self.assertEqual(get_linesep(output_dest), "\n")

    def test_get_output_destination_with_path_input_returns_writable_io(
        self,
    ):

        """Test that by calling get_output_destination with a Path as input
        we obtain write access to the file of Path."""
        mocked_open = mock_open()
        with patch("builtins.open", mocked_open):
            with get_output_destination(pathlib.Path("test")) as destination:
                self.assertTrue(
                    destination.writable(),
                    msg="file destination is writable",
                )
        mocked_open.assert_called_once_with(pathlib.Path("test"), "w")

    def test_get_output_destination_without_input_returns_stdout(
        self,
    ):

        """Test that by calling get_output_destination without input
        we obtain sys.stdout."""
        with get_output_destination() as destination:
            self.assertEqual(
                destination,
                sys.stdout,
                msg="default destination is sys.stdout",
            )

    def test_closing_get_output_destination_does_not_close_stdout(
        self,
    ):

        """Test that get_output_destination() can be safely used without closing sys.stdout."""

        with get_output_destination() as _:
            pass
        output_catcher = StringIO()
        with contextlib.redirect_stdout(output_catcher):
            sys.stdout.write("test after context closed")
        self.assertEqual(
            output_catcher.getvalue(),
            "test after context closed",
            msg="Can write to sys.stdout after closing desitnation context",
        )

    def test_get_header_for_csv(
        self,
    ):

        """Test that by calling get_header with input csv we get the correct line."""

        header = get_header("linesep", "csv")
        self.assertEqual(
            header,
            "File,Rg,Rg StDev,I(0),I(0) StDev,First point,Last point,Quality,Aggregatedlinesep",
            msg="csv header is correct",
        )

    def test_get_header_for_ssv(
        self,
    ):

        """Test that by calling get_header with input ssv we get an empty string."""

        header = get_header("linesep", "ssv")
        self.assertEqual(
            header,
            "",
            msg="ssv header is correct",
        )

    def test_get_header_for_native(
        self,
    ):

        """Test that by calling get_header with input native we get an empty string."""

        header = get_header("linesep", "native")
        self.assertEqual(
            header,
            "",
            msg="native header is correct",
        )

    def test_get_header_without_input_format(
        self,
    ):

        """Test that by calling get_header without input format we get an empty string."""

        header = get_header("linesep", None)
        self.assertEqual(
            header,
            "",
            msg="header for undefined format is correct",
        )

    def test_collect_files_only_returns_existing_files(self):

        """Test that collect_files discards strings that do not match an existing file."""

        def os_stat_mock(path):
            if "good" in pathlib.Path(path).name:
                pass
            else:
                if sys.version_info.minor > 7:
                    raise ValueError
                else:
                    raise OSError(ENOENT, "dummy")

        mocked_stat = MagicMock(side_effect=os_stat_mock)
        with patch("os.stat", mocked_stat):
            local_pathlib = importlib.import_module("pathlib")
            local_pathlib = importlib.reload(local_pathlib)
            fit = importlib.import_module("..fitting", "freesas.subpkg")
            fit = importlib.reload(fit)
            self.assertEqual(
                fit.collect_files(["testgood", "testbad"]),
                [local_pathlib.Path("testgood")],
            )
        # Reload without the patch
        local_pathlib = importlib.reload(local_pathlib)
        reload_os_and_fitting()

    @patch("platform.system", MagicMock(return_value="Windows"))
    def test_collect_files_globs_on_windows(self):

        """Test that collect_files globs on Windows if no existent files provided."""

        def os_stat_mock(path):
            if sys.version_info.minor > 7:
                raise ValueError
            else:
                raise OSError(ENOENT, "dummy")

        mocked_stat = MagicMock(side_effect=os_stat_mock)
        mocked_glob = MagicMock(
            side_effect=[
                (p for p in [pathlib.Path("pathA"), pathlib.Path("pathB")])
            ]
        )
        with patch("os.stat", mocked_stat):
            with patch.object(pathlib.Path, "glob", mocked_glob):
                fit = importlib.import_module("..fitting", "freesas.subpkg")
                fit = importlib.reload(fit)
                self.assertEqual(
                    fit.collect_files(["testgood"]),
                    [pathlib.Path("pathA"), pathlib.Path("pathB")],
                    msg="collect_files on windows returns list if fiel argument does not exist",
                )
        mocked_glob.assert_called_once()

        # Reload without the patch
        reload_os_and_fitting()

    def test_rg_result_line_csv(self):

        """Test the formatting of a csv result line for  a Guinier fit."""

        test_result = RG_RESULT(3.1, 0.1, 103, 2.5, 13, 207, 50.1, 0.05)
        expected_line = "test.file,3.1000,0.1000,103.0000,2.5000, 13,207,50.1000,0.0500lineend"
        obtained_line = rg_result_to_output_line(
            rg_result=test_result,
            afile=pathlib.Path("test.file"),
            output_format="csv",
            linesep="lineend",
        )
        self.assertEqual(
            obtained_line, expected_line, msg="csv line for RG_Result correct"
        )

    def test_rg_result_line_ssv(self):

        """Test the formatting of a ssv result line for  a Guinier fit."""

        test_result = RG_RESULT(3.1, 0.1, 103, 2.5, 13, 207, 50.1, 0.05)
        expected_line = "3.1000 0.1000 103.0000 2.5000  13 207 50.1000 0.0500 test.filelineend"
        obtained_line = rg_result_to_output_line(
            rg_result=test_result,
            afile=pathlib.Path("test.file"),
            output_format="ssv",
            linesep="lineend",
        )
        self.assertEqual(
            obtained_line, expected_line, msg="ssv line for RG_Result correct"
        )

    def test_rg_result_line_native(self):
        """Test the formatting of a native result line for  a Guinier fit."""
        test_result = RG_RESULT(3.1, 0.1, 103, 2.5, 13, 207, 50.1, 0.05)
        expected_line = "test.file Rg=3.1000(±0.1000) I0=103.0000(±2.5000) [13-207] 5010.00% lineend"
        obtained_line = rg_result_to_output_line(
            rg_result=test_result,
            afile=pathlib.Path("test.file"),
            output_format="native",
            linesep="lineend",
        )
        self.assertEqual(
            obtained_line,
            expected_line,
            msg="native line for RG_Result correct",
        )

    def test_rg_result_line_no_format(self):

        """Test the formatting of a native result line for  a Guinier fit."""

        test_result = RG_RESULT(3.1, 0.1, 103, 2.5, 13, 207, 50.1, 0.05)
        expected_line = "test.file Rg=3.1000(±0.1000) I0=103.0000(±2.5000) [13-207] 5010.00% lineend"
        obtained_line = rg_result_to_output_line(
            rg_result=test_result,
            afile=pathlib.Path("test.file"),
            linesep="lineend",
        )
        self.assertEqual(
            obtained_line,
            expected_line,
            msg="line for RG_Result without format specification correct",
        )

    @patch(
        "freesas.fitting.collect_files",
        MagicMock(return_value=[pathlib.Path("test")]),
    )
    @patch(
        "freesas.fitting.load_scattering_data",
        MagicMock(
            return_value=numpy.array(
                [[1.0, 1.0, 1.0], [2.0, 2.0, 1.0], [3.0, 3.0, 3.0]]
            )
        ),
    )
    @patch(
        "freesas.fitting.get_linesep",
        MagicMock(return_value="linesep"),
    )
    def test_run_guinier_fit_uses_provided_fit_function(self):
        """Test that run_guinier_fit uses fit function provided in the arguments
        and outputs its result in a line."""

        @counted
        def dummy_fit_function(input_data: numpy.ndarray) -> RG_RESULT:
            return RG_RESULT(3.1, 0.1, 103, 2.5, 13, 207, 50.1, 0.05)

        dummy_parser = get_dummy_guinier_parser(
            verbose=0,
            file=[pathlib.Path("test")],
            output=None,
            format=None,
            unit="nm",
        )
        output_catcher = StringIO()
        with contextlib.redirect_stdout(output_catcher):
            run_guinier_fit(
                fit_function=dummy_fit_function,
                parser=dummy_parser,
                logger=logger,
            )
        expected_output = "test Rg=3.1000(±0.1000) I0=103.0000(±2.5000) [13-207] 5010.00% linesep"
        self.assertEqual(
            output_catcher.getvalue(),
            expected_output,
            msg="run_guinier_fit provides expected output",
        )
        self.assertEqual(
            dummy_fit_function.calls,
            1,
            msg="Provided fit function was called once",
        )

    @patch(
        "freesas.fitting.collect_files",
        MagicMock(return_value=[pathlib.Path("test"), pathlib.Path("test")]),
    )
    @patch(
        "freesas.fitting.load_scattering_data",
        MagicMock(
            return_value=numpy.array(
                [[1.0, 1.0, 1.0], [2.0, 2.0, 1.0], [3.0, 3.0, 3.0]]
            )
        ),
    )
    @patch(
        "freesas.fitting.get_linesep",
        MagicMock(return_value="linesep"),
    )
    def test_run_guinier_fit_iterates_over_files(self):

        """Test that run_guinier_fit calls the provided fit function for each provided file."""

        @counted
        def dummy_fit_function(input_data: numpy.ndarray) -> RG_RESULT:
            return RG_RESULT(3.1, 0.1, 103, 2.5, 13, 207, 50.1, 0.05)

        dummy_parser = get_dummy_guinier_parser(
            verbose=0,
            file=[pathlib.Path("test"), pathlib.Path("test")],
            output=None,
            format=None,
            unit="nm",
        )
        output_catcher = StringIO()
        with contextlib.redirect_stdout(output_catcher):
            run_guinier_fit(
                fit_function=dummy_fit_function,
                parser=dummy_parser,
                logger=logger,
            )
        expected_output = (
            "test Rg=3.1000(±0.1000) I0=103.0000(±2.5000) [13-207] 5010.00% linesep"
            * 2
        )
        self.assertEqual(
            output_catcher.getvalue(),
            expected_output,
            msg="run_guinier_fit provides expected output",
        )
        self.assertEqual(
            dummy_fit_function.calls,
            2,
            msg="Provided fit function was called twice",
        )

    @unittest.skip("Unreliable")
    @patch(
        "freesas.fitting.collect_files",
        MagicMock(return_value=[pathlib.Path("test"), pathlib.Path("test2")]),
    )
    @patch(
        "freesas.fitting.load_scattering_data",
        MagicMock(
            side_effect=build_mock_for_load_scattering_with_Errors(
                {pathlib.Path("test").name: OSError}
            )
        ),
    )
    @patch(
        "freesas.fitting.get_linesep",
        MagicMock(return_value="linesep"),
    )
    def test_run_guinier_outputs_error_if_file_not_found(self):
        """Test that run_guinier_fit outputs an error if data loading raises OSError
        and continues to the next file."""

        @counted
        def dummy_fit_function(input_data: numpy.ndarray) -> RG_RESULT:
            return RG_RESULT(3.1, 0.1, 103, 2.5, 13, 207, 50.1, 0.05)

        dummy_parser = get_dummy_guinier_parser(
            verbose=0,
            file=[pathlib.Path("test"), pathlib.Path("test2")],
            output=None,
            format=None,
            unit="nm",
        )
        output_catcher_stdout = StringIO()
        output_catcher_stderr = StringIO()
        with contextlib.redirect_stdout(output_catcher_stdout):
            with contextlib.redirect_stderr(output_catcher_stderr):
                run_guinier_fit(
                    fit_function=dummy_fit_function,
                    parser=dummy_parser,
                    logger=logger,
                )
        expected_stdout_output = "test2 Rg=3.1000(±0.1000) I0=103.0000(±2.5000) [13-207] 5010.00% linesep"
        self.assertEqual(
            output_catcher_stdout.getvalue(),
            expected_stdout_output,
            msg="run_guinier_fit provides expected stdout output",
        )
        expected_stderr_output = (
            "ERROR:freesas.test.test_fitting:Unable to read file test"
        )
        self.assertTrue(
            expected_stderr_output in output_catcher_stderr.getvalue(),
            msg="run_guinier_fit provides expected stderr output",
        )
        self.assertEqual(
            dummy_fit_function.calls,
            1,
            msg="Provided fit function was called once",
        )

    @unittest.skip("Unreliable")
    @patch(
        "freesas.fitting.load_scattering_data",
        MagicMock(
            side_effect=build_mock_for_load_scattering_with_Errors(
                {pathlib.Path("test").name: ValueError}
            )
        ),
    )
    @patch_collect_files
    @patch_linesep
    def test_run_guinier_outputs_error_if_file_not_parsable(self):

        """Test that run_guinier_fit outputs an error if data loading raises ValueError
        and continues to the next file."""

        @counted
        def dummy_fit_function(input_data: numpy.ndarray) -> RG_RESULT:
            return RG_RESULT(3.1, 0.1, 103, 2.5, 13, 207, 50.1, 0.05)

        dummy_parser = get_dummy_guinier_parser(
            verbose=0,
            file=[pathlib.Path("test"), pathlib.Path("test2")],
            output=None,
            format=None,
            unit="nm",
        )
        output_catcher_stdout = StringIO()
        output_catcher_stderr = StringIO()
        with contextlib.redirect_stdout(output_catcher_stdout):
            with contextlib.redirect_stderr(output_catcher_stderr):
                run_guinier_fit(
                    fit_function=dummy_fit_function,
                    parser=dummy_parser,
                    logger=logger,
                )
        expected_stdout_output = "test2 Rg=3.1000(±0.1000) I0=103.0000(±2.5000) [13-207] 5010.00% linesep"
        self.assertEqual(
            output_catcher_stdout.getvalue(),
            expected_stdout_output,
            msg="run_guinier_fit provides expected stdout output",
        )
        expected_stderr_output = (
            "ERROR:freesas.test.test_fitting:Unable to parse file test"
        )
        self.assertTrue(
            expected_stderr_output in output_catcher_stderr.getvalue(),
            msg="run_guinier_fit provides expected stderr output",
        )
        self.assertEqual(
            dummy_fit_function.calls,
            1,
            msg="Provided fit function was called once",
        )

    @patch(
        "freesas.fitting.load_scattering_data",
        MagicMock(
            side_effect=[
                numpy.array(
                    [[0.0, 1.0, 1.0], [2.0, 2.0, 1.0], [3.0, 3.0, 3.0]]
                ),
                numpy.array(
                    [[2.0, 1.0, 1.0], [2.0, 2.0, 1.0], [3.0, 3.0, 3.0]]
                ),
            ]
        ),
    )
    @patch_collect_files
    @patch_linesep
    def test_run_guinier_outputs_error_if_fitting_raises_insufficient_data_error(
        self,
    ):

        """Test that run_guinier_fit outputs an error if fitting raises InsufficientDataError
        and continues to the next file."""

        @counted
        def dummy_fit_function(input_data: numpy.ndarray) -> RG_RESULT:
            if input_data[0, 0] <= 0.1:
                raise InsufficientDataError
            return RG_RESULT(3.1, 0.1, 103, 2.5, 13, 207, 50.1, 0.05)

        dummy_parser = get_dummy_guinier_parser(
            verbose=0,
            file=[pathlib.Path("test"), pathlib.Path("test2")],
            output=None,
            format=None,
            unit="nm",
        )
        output_catcher_stdout = StringIO()
        output_catcher_stderr = StringIO()
        with contextlib.redirect_stdout(output_catcher_stdout):
            with contextlib.redirect_stderr(output_catcher_stderr):
                run_guinier_fit(
                    fit_function=dummy_fit_function,
                    parser=dummy_parser,
                    logger=logger,
                )
        expected_stdout_output = "test2 Rg=3.1000(±0.1000) I0=103.0000(±2.5000) [13-207] 5010.00% linesep"
        self.assertEqual(
            output_catcher_stdout.getvalue(),
            expected_stdout_output,
            msg="run_guinier_fit provides expected stdout output",
        )
        expected_stderr_output = "test, InsufficientDataError:  "
        self.assertTrue(
            expected_stderr_output in output_catcher_stderr.getvalue(),
            msg="run_guinier_fit provides expected stderr output for fitting InsufficientError",
        )
        self.assertEqual(
            dummy_fit_function.calls,
            2,
            msg="Provided fit function was called twice",
        )

    @patch(
        "freesas.fitting.load_scattering_data",
        MagicMock(
            side_effect=[
                numpy.array(
                    [[0.0, 1.0, 1.0], [2.0, 2.0, 1.0], [3.0, 3.0, 3.0]]
                ),
                numpy.array(
                    [[2.0, 1.0, 1.0], [2.0, 2.0, 1.0], [3.0, 3.0, 3.0]]
                ),
            ]
        ),
    )
    @patch_collect_files
    @patch_linesep
    def test_run_guinier_outputs_error_if_fitting_raises_no_guinier_region_error(
        self,
    ):

        """Test that run_guinier_fit outputs an error if fitting raises NoGuinierRegionError
        and continues to the next file."""

        @counted
        def dummy_fit_function(input_data: numpy.ndarray) -> RG_RESULT:
            if input_data[0, 0] <= 0.1:
                raise NoGuinierRegionError
            return RG_RESULT(3.1, 0.1, 103, 2.5, 13, 207, 50.1, 0.05)

        dummy_parser = get_dummy_guinier_parser(
            verbose=0,
            file=[pathlib.Path("test"), pathlib.Path("test2")],
            output=None,
            format=None,
            unit="nm",
        )
        output_catcher_stdout = StringIO()
        output_catcher_stderr = StringIO()
        with contextlib.redirect_stdout(output_catcher_stdout):
            with contextlib.redirect_stderr(output_catcher_stderr):
                run_guinier_fit(
                    fit_function=dummy_fit_function,
                    parser=dummy_parser,
                    logger=logger,
                )
        expected_stdout_output = "test2 Rg=3.1000(±0.1000) I0=103.0000(±2.5000) [13-207] 5010.00% linesep"
        self.assertEqual(
            output_catcher_stdout.getvalue(),
            expected_stdout_output,
            msg="run_guinier_fit provides expected stdout output",
        )
        expected_stderr_output = "test, NoGuinierRegionError:  "
        self.assertTrue(
            expected_stderr_output in output_catcher_stderr.getvalue(),
            msg="run_guinier_fit provides expected stderr output fitting NoGuinierRegionError",
        )
        self.assertEqual(
            dummy_fit_function.calls,
            2,
            msg="Provided fit function was called twice",
        )

    @patch(
        "freesas.fitting.load_scattering_data",
        MagicMock(
            side_effect=[
                numpy.array(
                    [[0.0, 1.0, 1.0], [2.0, 2.0, 1.0], [3.0, 3.0, 3.0]]
                ),
                numpy.array(
                    [[2.0, 1.0, 1.0], [2.0, 2.0, 1.0], [3.0, 3.0, 3.0]]
                ),
            ]
        ),
    )
    @patch_collect_files
    @patch_linesep
    def test_run_guinier_outputs_error_if_fitting_raises_value_error(
        self,
    ):

        """Test that run_guinier_fit outputs an error if fitting raises ValueError
        and continues to the next file."""

        @counted
        def dummy_fit_function(input_data: numpy.ndarray) -> RG_RESULT:
            if input_data[0, 0] <= 0.1:
                raise ValueError
            return RG_RESULT(3.1, 0.1, 103, 2.5, 13, 207, 50.1, 0.05)

        dummy_parser = get_dummy_guinier_parser(
            verbose=0,
            file=[pathlib.Path("test"), pathlib.Path("test2")],
            output=None,
            format=None,
            unit="nm",
        )
        output_catcher_stdout = StringIO()
        output_catcher_stderr = StringIO()
        with contextlib.redirect_stdout(output_catcher_stdout):
            with contextlib.redirect_stderr(output_catcher_stderr):
                run_guinier_fit(
                    fit_function=dummy_fit_function,
                    parser=dummy_parser,
                    logger=logger,
                )
        expected_stdout_output = "test2 Rg=3.1000(±0.1000) I0=103.0000(±2.5000) [13-207] 5010.00% linesep"
        self.assertEqual(
            output_catcher_stdout.getvalue(),
            expected_stdout_output,
            msg="run_guinier_fit provides expected stdout output",
        )
        expected_stderr_output = "test, ValueError: "
        self.assertTrue(
            expected_stderr_output in output_catcher_stderr.getvalue(),
            msg="run_guinier_fit provides expected stderr output fitting ValueError",
        )
        self.assertEqual(
            dummy_fit_function.calls,
            2,
            msg="Provided fit function was called twice",
        )

    @patch(
        "freesas.fitting.load_scattering_data",
        MagicMock(
            side_effect=[
                numpy.array(
                    [[0.0, 1.0, 1.0], [2.0, 2.0, 1.0], [3.0, 3.0, 3.0]]
                ),
                numpy.array(
                    [[2.0, 1.0, 1.0], [2.0, 2.0, 1.0], [3.0, 3.0, 3.0]]
                ),
            ]
        ),
    )
    @patch_collect_files
    @patch_linesep
    def test_run_guinier_outputs_error_if_fitting_raises_index_error(
        self,
    ):

        """Test that run_guinier_fit outputs an error if fitting raises IndexError
        and continues to the next file."""

        @counted
        def dummy_fit_function(input_data: numpy.ndarray) -> RG_RESULT:
            if input_data[0, 0] <= 0.1:
                raise IndexError
            return RG_RESULT(3.1, 0.1, 103, 2.5, 13, 207, 50.1, 0.05)

        dummy_parser = get_dummy_guinier_parser(
            verbose=0,
            file=[pathlib.Path("test"), pathlib.Path("test2")],
            output=None,
            format=None,
            unit="nm",
        )
        output_catcher_stdout = StringIO()
        output_catcher_stderr = StringIO()
        with contextlib.redirect_stdout(output_catcher_stdout):
            with contextlib.redirect_stderr(output_catcher_stderr):
                run_guinier_fit(
                    fit_function=dummy_fit_function,
                    parser=dummy_parser,
                    logger=logger,
                )
        expected_stdout_output = "test2 Rg=3.1000(±0.1000) I0=103.0000(±2.5000) [13-207] 5010.00% linesep"
        self.assertEqual(
            output_catcher_stdout.getvalue(),
            expected_stdout_output,
            msg="run_guinier_fit provides expected stdout output",
        )
        expected_stderr_output = "test, IndexError: "
        self.assertTrue(
            expected_stderr_output in output_catcher_stderr.getvalue(),
            msg="run_guinier_fit provides expected stderr output for fitting IndexError",
        )
        self.assertEqual(
            dummy_fit_function.calls,
            2,
            msg="Provided fit function was called twice",
        )


def suite():
    """Build a test suite from the TestFitting class."""

    test_suite = unittest.TestSuite()
    for class_element in dir(TestFitting):
        if class_element.startswith("test"):
            test_suite.addTest(TestFitting(class_element))
    return test_suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



# -*- coding: utf-8 -*-
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2017  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__authors__ = ["Martha Brennich"]
__license__ = "MIT"
__date__ = "16/07/2021"

import unittest
import pathlib
import re
import logging
from subprocess import run, Popen, PIPE, STDOUT
from os import linesep
import PyPDF2
from freesas.test.utilstests import get_datafile
logger = logging.getLogger(__name__)

expectedTexts = {
    "Label of scatter plot X-axis": r"\$q\$ \(nm\$\^\{-1\}\$\)",
    "Label of scatter plot Y-axis": r"\$I\(q\)\$ \(log scale\)",
    "Smallest Tick of scatter plot Y axis": r"\$\\mathdefault\{10\^\{-2\}\}\$",
    "Largest Tick of scatter plot Y axis": r"\$\\mathdefault\{10\^\{2\}\}\$",
    "Scattering plot caption": r"Scattering curve",
    "Experimental data legend": r"Experimental data",  # used twice, but we might just ignore that
    "Guinier region legend": r"Guinier region: \$R_g=\$[23]\.[0-9][0-9] nm, \$I_0=\$6[0-9]\.[0-9][0-9]",
    "BIFT fit legend": r"BIFT extraplolated: \$D_\{max\}=\$9\.[0-9][0-9] nm",
    "Label of Guinier plot X-axis": r"\$q\^2\$ \(nm\$\^\{-2\}\$\)",
    "Label of Guinier plot Y-axis": r"ln\[\$I\(q\)\$]",
    "Guinier - qRgmin": r"\$\(qR_\{g\}\)_\{min\}\$=0\.[0-9]",
    "Guinier - qRgmax": r"\$\(qR_\{g\}\)_\{max\}\$=1\.[0123]",
    "Guinier region label": r"Guinier region",
    "Guinier plot caption": r"Guinier plot: \$R_\{g\}=\$[23]\.[0-9][0-9] nm \$I_\{0\}=\$6[0-9]\.[0-9][0-9]",
    "Guinier fit equation": r"ln\[\$I\(q\)\$\] = 4\.12 -3\.01 \* \$q\^2\$",
    "Guinier fit data label": r"Experimental curve",
    "Label of Kratky plot X-Axis": r"\$qR_\{g\}\$",
    "Label of Kratky plot Y-Axis": r"\$\(qR_\{g\}\)\^2 I/I_\{0\}\$",
    "Kratky plot caption": r"Dimensionless Kratky plot",
    "Label of distribution plot X-axis": r"\$r\$ \(nm\)",
    "Label of distribution plot Y-axis": r"\$p\(r\)\$",
    "Distribution plot caption": r"Pair distribution function",
    # "BIFT chi": r"BIFT: χ\$_\{r\}\^\{2\}=\$1\.[0-9][0-9]",
    "BIFT Dmax": r"\$D_\{max\}=\$[1]?[09].[0-9][0-9] nm",
    "BIFT Rg": r"\$R_\{g\}=\$[23]\.[0-9][0-9] nm",
    "BIFT I0": r"\$I_\{0\}=\$6[0-9]\.[0-9][0-9]",
}


class TestFreeSAS(unittest.TestCase):

    cwd = pathlib.Path.cwd()
    TEST_IMAGE_NAME = pathlib.Path(cwd, "freesas.svg")
    TEST_PDF_NAME = pathlib.Path(cwd, "freesas.pdf")
    test_location = pathlib.Path(__file__)
    test_data_location = pathlib.Path(test_location.parent, "e2etest_data")
    bsa_filename = pathlib.Path(get_datafile("bsa_005_sub.dat"))
    sas_curve2_filename = pathlib.Path(get_datafile("SASDF52.dat"))
    image_text = None

    @classmethod
    def setUpClass(cls):
        super(TestFreeSAS, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        super(TestFreeSAS, cls).tearDownClass()
        cls.image_text = None

    def __init__(self, testName, **extra_kwargs):
        super(TestFreeSAS, self).__init__(testName)
        self.extra_arg = extra_kwargs

    def test_save_image(self):
        """
        Test whether freeSAS finishes without errors
        if there is an -o argument.
        It also uses the output as input for label tests.
        """
        # Make sure the result file does not exist for a meaningful assert
        try:
            self.TEST_IMAGE_NAME.unlink()
        except FileNotFoundError:
            pass
        run_freesas = run(
            [
                "freesas",
                str(self.bsa_filename),
                "-o",
                str(self.TEST_IMAGE_NAME),
            ],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )
        self.assertEqual(
            run_freesas.returncode, 0, msg="freesas completed well"
        )
        self.assertTrue(self.TEST_IMAGE_NAME.exists(), msg="Found output file")
        with open(self.TEST_IMAGE_NAME) as file:
            self.__class__.image_text = file.read()
        try:
            self.TEST_IMAGE_NAME.unlink()
        except FileNotFoundError:
            pass

    def test_display_image(self):
        """
        Test whether freeSAS for one dataset finishes without errors
        if there no -o argument.
        """
        run_freesas = Popen(
            ["freesas", str(self.bsa_filename)],
            universal_newlines=True,
            stdout=PIPE,
            stderr=PIPE,
            stdin=PIPE,
        )
        stdout, _ = run_freesas.communicate(linesep, timeout=40)
        self.assertEqual(
            run_freesas.returncode, 0, msg="freesas completed well"
        )
        self.assertEqual(
            stdout, "Press enter to quit", msg="freesas requested enter"
        )

    def test_label(self):
        """
        Test for the presence of labels in the svg.
        Requires two extra kwargs:
        regex: The regex expression to search for
        description: The description of what the label represents
        """
        text_regex: str = self.extra_arg["regex"]
        text_description: str = self.extra_arg["description"]
        pattern = re.compile(text_regex)
        self.assertIsNotNone(
            pattern.search(self.image_text),
            msg="Could not find text for {} in image".format(text_description),
        )

    def test_multi_file_pdf(self):
        """
        Check that correct PDF is created when processing several files with the -o option.
        """
        # Make sure the result file does not exist for a meaningful assert
        try:
            self.TEST_PDF_NAME.unlink()
        except FileNotFoundError:
            pass
        run_freesas = run(
            [
                "freesas",
                str(self.bsa_filename),
                str(self.sas_curve2_filename),
                "-o",
                str(self.TEST_PDF_NAME),
            ],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )
        self.assertEqual(
            run_freesas.returncode, 0, msg="freesas completed well"
        )
        self.assertTrue(self.TEST_PDF_NAME.exists(), msg="Found output file")
        with open(self.TEST_PDF_NAME, "rb") as file:
            output_pdf = PyPDF2.PdfFileReader(file)
            self.assertEqual(
                output_pdf.numPages, 2, msg="correct number of pages in pdf"
            )
            page_1_text = output_pdf.getPage(0).extractText()
            page_2_text = output_pdf.getPage(1).extractText()
        self.assertTrue(
            (str(self.bsa_filename) in page_1_text)
            ^ (str(self.bsa_filename) in page_2_text),
            msg=str(self.bsa_filename) + " found on one of the pages",
        )
        self.assertTrue(
            (str(self.sas_curve2_filename) in page_1_text)
            ^ (str(self.sas_curve2_filename) in page_2_text),
            msg=str(self.sas_curve2_filename) + " found on one of the pages",
        )
        # Clean up
        try:
            self.TEST_PDF_NAME.unlink()
        except FileNotFoundError:
            pass


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(TestFreeSAS("test_display_image"))
    test_suite.addTest(TestFreeSAS("test_save_image"))
    for text_description, text_regex in expectedTexts.items():
        test_suite.addTest(
            TestFreeSAS(
                "test_label", regex=text_regex, description=text_description
            )
        )
    test_suite.addTest(TestFreeSAS("test_multi_file_pdf"))
    return test_suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



"""End to end tests for auto_gpa.py """

__authors__ = ["Martha Brennich"]
__license__ = "MIT"
__date__ = "16/07/2021"

import unittest
import pathlib
import logging
from platform import system
from subprocess import run, PIPE, STDOUT
from os import linesep
from os.path import normpath
import codecs
import parse
from numpy import loadtxt
from freesas.test.utilstests import get_datafile

logger = logging.getLogger(__name__)

if system() == "Windows":
    free_bift = "free_bift.exe"
else:
    free_bift = "free_bift"


class TestBIFT(unittest.TestCase):
    """End to end tests for free_bift"""

    cwd = pathlib.Path.cwd()
    test_location = pathlib.Path(__file__)
    test_data_location = pathlib.Path(test_location.parent, "e2etest_data")
    bsa_filename = pathlib.Path(get_datafile("bsa_005_sub.dat"))
    sas_curve2_filename = pathlib.Path(get_datafile("SASDF52.dat"))
    SASDFX7 = pathlib.Path(get_datafile("SASDFX7.dat"))
    expected_outfile_name_bsa = pathlib.Path(
        cwd, bsa_filename.name
    ).with_suffix(".out")

    def __init__(self, testName, **extra_kwargs):
        super().__init__(testName)
        self.extra_arg = extra_kwargs

    def remove_output_files(self):
        try:
            self.expected_outfile_name_bsa.unlink()
        except FileNotFoundError:
            pass

    def setUp(self):
        self.remove_output_files()
        return super().setUp()

    def tearDown(self):
        self.remove_output_files()
        return super().tearDown()

    def test_bm29_bsa_without_arguments_creates_out_file(self):
        """
        Test whether bift app on BSA data from BM29 creates an out file.
        """

        run_app = run(
            [free_bift, normpath(str(self.bsa_filename))],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )
        self.assertEqual(
            run_app.returncode, 0, msg="bift on BM29 BSA completed well"
        )
        self.assertTrue(
            self.expected_outfile_name_bsa.exists(),
            f"bift on BM29 BSA created out file with correct name: {str(self.expected_outfile_name_bsa)}",
        )

    def test_bm29_bsa_out_file_has_the_expected_format(self):
        """
        Test whether bift app on BSA data from BM29 creates an out file.
        """

        _ = run(
            [free_bift, normpath(str(self.bsa_filename))],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )

        with open(
            self.expected_outfile_name_bsa, "r", encoding="utf-8"
        ) as out_file:
            out_file_content = out_file.readlines()
        with codecs.open(
            self.expected_outfile_name_bsa, encoding="utf-8"
        ) as filecp:
            data = loadtxt(
                filecp,
                dtype=float,
                delimiter="\t",
                skiprows=9,
            )

        self.assertEqual(out_file_content[0].strip(), f"# {self.bsa_filename}")
        self.assertTrue(
            out_file_content[1].startswith("# Dmax= ")
            and "±" in out_file_content[1],
            msg=f"exptexted line to resemble # Dmax= 9.76±0.05 got {out_file_content[1]}",
        )
        self.assertTrue(
            out_file_content[2].startswith("# 𝛂= ")
            and "±" in out_file_content[2],
            msg=f"exptexted line to resemble # 𝛂= 8764.5±1384.2 got {out_file_content[2]}",
        )
        self.assertTrue(
            out_file_content[3].startswith("# S₀= ")
            and "±" in out_file_content[3],
            msg=f"exptexted line to resemble # S₀= 0.0002±0.0000  got {out_file_content[3]}",
        )
        self.assertTrue(
            out_file_content[4].startswith("# χ²= ")
            and "±" in out_file_content[4],
            msg=f"exptexted line to resemble # χ²= 1.89±0.00 got {out_file_content[4]}",
        )
        self.assertTrue(
            out_file_content[5].startswith("# logP= ")
            and "±" in out_file_content[5],
            msg=f"exptexted line to resemble # logP= -914.15±0.47 got {out_file_content[5]}",
        )
        self.assertTrue(
            out_file_content[6].startswith("# Rg= ")
            and "±" in out_file_content[6],
            msg=f"exptexted line to resemble # Rg= 2.98±0.00  got {out_file_content[6]}",
        )
        self.assertTrue(
            out_file_content[7].startswith("# I₀= ")
            and "±" in out_file_content[7],
            msg=f"exptexted line to resemble 60.86±0.00 got {out_file_content[7]}",
        )
        self.assertEqual(out_file_content[8].strip(), "")
        self.assertEqual(
            out_file_content[9].strip(),
            "# r\tp(r)\tsigma_p(r)",
        )
        self.assertEqual(
            data.shape[1],
            3,
        )

    def test_bm29_bsa_result_numerically_matches_expectations(self):
        """
        Test whether the results of the bift app on BM29 BSA give roughly the
        expected Dmax, I₀ anr Rg and that the first is and the last point is close to 0.
        """

        run_app_ = run(
            [free_bift, normpath(str(self.bsa_filename))],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )

        with open(
            self.expected_outfile_name_bsa, "r", encoding="utf-8"
        ) as out_file:
            out_file_content = out_file.readlines()

        self.assertAlmostEqual(
            float(out_file_content[1][8:12]),
            9.75,
            places=1,
            msg=f"expected Dmax to be close to 0.75 got {out_file_content[1]}",
        )

        self.assertAlmostEqual(
            float(out_file_content[6][6:10]),
            3.0,
            places=1,
            msg=f"expected Rg to be close to 3.0 got {out_file_content[6]}",
        )

        self.assertAlmostEqual(
            0.1 * float(out_file_content[7][6:10]),
            6.1,
            places=1,
            msg=f"expected I0 to be close to 60 got {out_file_content[7]}",
        )

        self.assertEqual(
            out_file_content[10].strip(),
            "0.0\t0.0\t0.0",
            msg=f"Expected first p(r) line to be '0.0     0.0     0.0' got {out_file_content[10]}",
        )

        last_line_content = out_file_content[-1].split("\t")

        self.assertAlmostEqual(
            float(last_line_content[0]),
            9.75,
            places=1,
            msg=f"expected last r point to be close to 9.75 got {last_line_content[0]}",
        )

        self.assertAlmostEqual(
            float(last_line_content[1]),
            0,
            places=2,
            msg=f"expected last r point to be close to 0 got {last_line_content[1]}",
        )

    def test_free_bift_outputs_one_line_summary(self):
        """
        Test whether free_bift app on BM29 BSA puts a one line summary in stdout.
        """

        run_app = run(
            [free_bift, normpath(str(self.bsa_filename))],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )

        if system() == "Windows":
            run_app_output = str(run_app.stdout, encoding="utf-16")[:-1].replace("\\\\", "\\")
        else:
            run_app_output = str(run_app.stdout, encoding="utf-8")[:-1]
        run_app_output_parsed = parse.parse(
            "bsa_005_sub.out: Dmax= {Dmax}±{Dmax_err}; 𝛂= {alpha}±{alpha_err}; S₀= {S0}±{S0_err}; χ²= {chi_squared}±{chi_squared_err}; logP= {logP}±{logP_err}; Rg= {Rg}±{Rg_err}; I₀= {I0}±{I0_err}",
            run_app_output,
        )
        self.assertListEqual(
            list(run_app_output_parsed.named),
            [
                "Dmax",
                "Dmax_err",
                "alpha",
                "alpha_err",
                "S0",
                "S0_err",
                "chi_squared",
                "chi_squared_err",
                "logP",
                "logP_err",
                "Rg",
                "Rg_err",
                "I0",
                "I0_err",
            ],
            msg="Could not parse free_bift std output",
        )

    def test_free_bift_values_of_one_line_summary_match_expectations(self):
        """
        Test whether the one line summary of free_bift app on BM29 BSA gives the expected values.
        """

        run_app = run(
            [free_bift, normpath(str(self.bsa_filename))],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )

        if system() == "Windows":
            run_app_output = str(run_app.stdout, encoding="utf-16")[:-1].replace("\\\\", "\\")
        else:
            run_app_output = str(run_app.stdout, encoding="utf-8")[:-1]
        run_app_output_parsed = parse.parse(
            "bsa_005_sub.out: Dmax= {Dmax}±{Dmax_err}; 𝛂= {alpha}±{alpha_err}; S₀= {S0}±{S0_err}; χ²= {chi_squared}±{chi_squared_err}; logP= {logP}±{logP_err}; Rg= {Rg}±{Rg_err}; I₀= {I0}±{I0_err}",
            run_app_output,
        )
        self.assertAlmostEqual(
            float(run_app_output_parsed["Dmax"]),
            9.75,
            places=1,
            msg=f"expected Dmax to be close to 0.75 got {run_app_output_parsed['Dmax']}",
        )

        self.assertAlmostEqual(
            float(run_app_output_parsed["Rg"]),
            3.0,
            places=1,
            msg=f"expected Rg to be close to 3.0 got {run_app_output_parsed['Rg']}",
        )

        self.assertAlmostEqual(
            0.1 * float(run_app_output_parsed["I0"]),
            6.1,
            places=1,
            msg=f"expected I0 to be close to 60 got {run_app_output_parsed['I0']}",
        )


def suite():
    """Build test suite for free_bift"""
    test_suite = unittest.TestSuite()

    test_suite.addTest(
        TestBIFT("test_bm29_bsa_without_arguments_creates_out_file")
    )
    test_suite.addTest(
        TestBIFT("test_bm29_bsa_out_file_has_the_expected_format")
    )
    test_suite.addTest(
        TestBIFT("test_bm29_bsa_result_numerically_matches_expectations")
    )
    test_suite.addTest(TestBIFT("test_free_bift_outputs_one_line_summary"))
    test_suite.addTest(
        TestBIFT(
            "test_free_bift_values_of_one_line_summary_match_expectations"
        )
    )
    return test_suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())
